import string
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Dict, Optional
from uuid import uuid4

import msgspec
import pytest
from hypothesis import given, settings
from hypothesis.strategies import dictionaries, integers, none, one_of, sampled_from, text, timedeltas

from litestar import Litestar, Request, Response, get
from litestar.security.jwt import JWTAuth, JWTCookieAuth, OAuth2PasswordBearerAuth, Token
from litestar.status_codes import HTTP_200_OK, HTTP_201_CREATED, HTTP_401_UNAUTHORIZED
from litestar.stores.memory import MemoryStore
from litestar.testing import create_test_client
from tests.models import User, UserFactory

if TYPE_CHECKING:
    from litestar.connection import ASGIConnection


@pytest.fixture(scope="module")
def mock_db() -> MemoryStore:
    return MemoryStore()


@given(
    algorithm=sampled_from(
        [
            "HS256",
            "HS384",
            "HS512",
        ]
    ),
    auth_header=sampled_from(["Authorization", "X-API-Key"]),
    default_token_expiration=timedeltas(min_value=timedelta(seconds=30), max_value=timedelta(weeks=1)),
    token_secret=text(min_size=10),
    response_status_code=integers(min_value=200, max_value=201),
    token_expiration=timedeltas(min_value=timedelta(seconds=30), max_value=timedelta(weeks=1)),
    token_issuer=one_of(none(), text(max_size=256)),
    token_audience=one_of(none(), text(max_size=256, alphabet=string.ascii_letters)),
    token_unique_jwt_id=one_of(none(), text(max_size=256)),
    token_extras=one_of(none(), dictionaries(text(max_size=256), text(max_size=256))),
)
@settings(deadline=None)
async def test_jwt_auth(
    mock_db: "MemoryStore",
    algorithm: str,
    auth_header: str,
    default_token_expiration: timedelta,
    token_secret: str,
    response_status_code: int,
    token_expiration: timedelta,
    token_issuer: Optional[str],
    token_audience: Optional[str],
    token_unique_jwt_id: Optional[str],
    token_extras: Optional[Dict[str, Any]],
) -> None:
    user = UserFactory.build()

    await mock_db.set(str(user.id), user, 120)  # type: ignore[arg-type]

    async def retrieve_user_handler(token: Token, _: "ASGIConnection") -> Any:
        return await mock_db.get(token.sub)

    jwt_auth = JWTAuth[Any](
        algorithm=algorithm,
        auth_header=auth_header,
        default_token_expiration=default_token_expiration,
        token_secret=token_secret,
        retrieve_user_handler=retrieve_user_handler,
    )

    @get("/my-endpoint", middleware=[jwt_auth.middleware])
    def my_handler(request: Request["User", Token, Any]) -> None:
        assert request.user
        assert msgspec.to_builtins(request.user) == msgspec.to_builtins(user)
        assert request.auth.sub == str(user.id)

    @get("/login")
    def login_handler() -> Response["User"]:
        return jwt_auth.login(
            identifier=str(user.id),
            response_body=user,
            response_status_code=response_status_code,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
            token_extras=token_extras,
        )

    with create_test_client(route_handlers=[my_handler, login_handler]) as client:
        response = client.get("/login")
        assert response.status_code == response_status_code
        _, _, encoded_token = response.headers.get(auth_header).partition(" ")
        assert encoded_token
        decoded_token = Token.decode(encoded_token=encoded_token, secret=token_secret, algorithm=algorithm)
        assert decoded_token.sub == str(user.id)
        assert decoded_token.iss == token_issuer
        assert decoded_token.aud == token_audience
        assert decoded_token.jti == token_unique_jwt_id
        if token_extras is not None:
            for key, value in token_extras.items():
                assert decoded_token.extras[key] == value

        response = client.get("/my-endpoint")
        assert response.status_code == HTTP_401_UNAUTHORIZED

        response = client.get("/my-endpoint", headers={auth_header: jwt_auth.format_auth_header(encoded_token)})
        assert response.status_code == HTTP_200_OK

        response = client.get("/my-endpoint", headers={auth_header: encoded_token})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        response = client.get("/my-endpoint", headers={auth_header: uuid4().hex})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        fake_token = Token(
            sub=uuid4().hex,
            iss=token_issuer,
            aud=token_audience,
            jti=token_unique_jwt_id,
            exp=(datetime.now(timezone.utc) + token_expiration),
        ).encode(secret=token_secret, algorithm=algorithm)

        response = client.get("/my-endpoint", headers={auth_header: jwt_auth.format_auth_header(fake_token)})
        assert response.status_code == HTTP_401_UNAUTHORIZED


@given(
    algorithm=sampled_from(
        [
            "HS256",
            "HS384",
            "HS512",
        ]
    ),
    auth_header=sampled_from(["Authorization", "X-API-Key"]),
    auth_cookie=sampled_from(["token", "accessToken"]),
    default_token_expiration=timedeltas(min_value=timedelta(seconds=30), max_value=timedelta(weeks=1)),
    token_secret=text(min_size=10),
    response_status_code=integers(min_value=200, max_value=201),
    token_expiration=timedeltas(min_value=timedelta(seconds=30), max_value=timedelta(weeks=1)),
    token_issuer=one_of(none(), text(max_size=256)),
    token_audience=one_of(none(), text(max_size=256, alphabet=string.ascii_letters)),
    token_unique_jwt_id=one_of(none(), text(max_size=256)),
    token_extras=one_of(none(), dictionaries(text(max_size=256), text(max_size=256))),
)
@settings(deadline=None)
async def test_jwt_cookie_auth(
    mock_db: "MemoryStore",
    algorithm: str,
    auth_header: str,
    auth_cookie: str,
    default_token_expiration: timedelta,
    token_secret: str,
    response_status_code: int,
    token_expiration: timedelta,
    token_issuer: Optional[str],
    token_audience: Optional[str],
    token_unique_jwt_id: Optional[str],
    token_extras: Optional[Dict[str, Any]],
) -> None:
    user = UserFactory.build()

    await mock_db.set(str(user.id), user, 120)  # type: ignore[arg-type]

    async def retrieve_user_handler(token: Token, connection: Any) -> Any:
        assert connection
        return await mock_db.get(token.sub)

    jwt_auth = JWTCookieAuth(
        algorithm=algorithm,
        key=auth_cookie,
        auth_header=auth_header,
        default_token_expiration=default_token_expiration,
        retrieve_user_handler=retrieve_user_handler,  # type: ignore[var-annotated]
        token_secret=token_secret,
    )

    @get("/my-endpoint", middleware=[jwt_auth.middleware])
    def my_handler(request: Request["User", Token, Any]) -> None:
        assert request.user
        assert msgspec.to_builtins(request.user) == msgspec.to_builtins(user)
        assert request.auth.sub == str(user.id)

    @get("/login")
    def login_handler() -> Response["User"]:
        return jwt_auth.login(
            identifier=str(user.id),
            response_body=user,
            response_status_code=response_status_code,
            token_expiration=token_expiration,
            token_issuer=token_issuer,
            token_audience=token_audience,
            token_unique_jwt_id=token_unique_jwt_id,
            token_extras=token_extras,
        )

    with create_test_client(route_handlers=[my_handler, login_handler]) as client:
        response = client.get("/login")
        assert response.status_code == response_status_code
        _, _, encoded_token = response.headers.get(auth_header).partition(" ")
        assert encoded_token
        decoded_token = Token.decode(encoded_token=encoded_token, secret=token_secret, algorithm=algorithm)
        assert decoded_token.sub == str(user.id)
        assert decoded_token.iss == token_issuer
        assert decoded_token.aud == token_audience
        assert decoded_token.jti == token_unique_jwt_id
        if token_extras is not None:
            for key, value in token_extras.items():
                assert decoded_token.extras[key] == value

        client.cookies.clear()
        response = client.get("/my-endpoint")
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies.clear()
        response = client.get("/my-endpoint", headers={auth_header: jwt_auth.format_auth_header(encoded_token)})
        assert response.status_code == HTTP_200_OK

        client.cookies = {auth_cookie: jwt_auth.format_auth_header(encoded_token)}  # type: ignore[assignment]
        response = client.get(
            "/my-endpoint",
        )
        assert response.status_code == HTTP_200_OK

        client.cookies.clear()
        response = client.get("/my-endpoint", headers={auth_header: encoded_token})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies.clear()
        response = client.get("/my-endpoint", headers={auth_cookie: encoded_token})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies.clear()
        response = client.get("/my-endpoint", headers={auth_header: jwt_auth.format_auth_header(uuid4().hex)})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies = {auth_cookie: jwt_auth.format_auth_header(uuid4().hex)}  # type: ignore[assignment]
        response = client.get("/my-endpoint")
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies = {auth_cookie: uuid4().hex}  # type: ignore[assignment]
        response = client.get("/my-endpoint")
        assert response.status_code == HTTP_401_UNAUTHORIZED
        fake_token = Token(
            sub=uuid4().hex,
            iss=token_issuer,
            aud=token_audience,
            jti=token_unique_jwt_id,
            exp=(datetime.now(timezone.utc) + token_expiration),
        ).encode(secret=token_secret, algorithm=algorithm)

        client.cookies.clear()
        response = client.get("/my-endpoint", headers={auth_header: jwt_auth.format_auth_header(fake_token)})
        assert response.status_code == HTTP_401_UNAUTHORIZED

        client.cookies = {auth_cookie: jwt_auth.format_auth_header(fake_token)}  # type: ignore[assignment]
        response = client.get("/my-endpoint")
        assert response.status_code == HTTP_401_UNAUTHORIZED


async def test_path_exclusion() -> None:
    async def retrieve_user_handler(_: Token, __: "ASGIConnection") -> None:
        return None

    jwt_auth = JWTAuth[Any](
        token_secret="abc123",
        retrieve_user_handler=retrieve_user_handler,
        exclude=["north", "south"],
    )

    @get("/north/{value:int}")
    def north_handler(value: int) -> Dict[str, int]:
        return {"value": value}

    @get("/south")
    def south_handler() -> None:
        return None

    @get("/west")
    def west_handler() -> None:
        return None

    with create_test_client(
        route_handlers=[north_handler, south_handler, west_handler], on_app_init=[jwt_auth.on_app_init]
    ) as client:
        response = client.get("/north/1")
        assert response.status_code == HTTP_200_OK

        response = client.get("/south")
        assert response.status_code == HTTP_200_OK

        response = client.get("/west")
        assert response.status_code == HTTP_401_UNAUTHORIZED


def test_jwt_auth_openapi() -> None:
    jwt_auth = JWTAuth[Any](token_secret="abc123", retrieve_user_handler=lambda _: None)  # type: ignore[arg-type, misc]
    assert jwt_auth.openapi_components.to_schema() == {
        "schemas": {},
        "securitySchemes": {
            "BearerToken": {
                "type": "http",
                "description": "JWT api-key authentication and authorization.",
                "name": "Authorization",
                "scheme": "Bearer",
                "bearerFormat": "JWT",
            }
        },
    }
    assert jwt_auth.security_requirement == {"BearerToken": []}
    app = Litestar(on_app_init=[jwt_auth.on_app_init])

    assert app.openapi_schema
    assert app.openapi_schema.to_schema() == {
        "openapi": "3.1.0",
        "info": {"title": "Litestar API", "version": "1.0.0"},
        "servers": [{"url": "/"}],
        "paths": {},
        "components": {
            "schemas": {},
            "securitySchemes": {
                "BearerToken": {
                    "type": "http",
                    "description": "JWT api-key authentication and authorization.",
                    "name": "Authorization",
                    "scheme": "Bearer",
                    "bearerFormat": "JWT",
                }
            },
        },
        "security": [{"BearerToken": []}],
    }


async def test_oauth2_password_bearer_auth_openapi(mock_db: "MemoryStore") -> None:
    user = UserFactory.build()

    await mock_db.set(str(user.id), user, 120)  # type: ignore[arg-type]

    async def retrieve_user_handler(token: Token, connection: Any) -> Any:
        assert connection
        return await mock_db.get(token.sub)

    jwt_auth = OAuth2PasswordBearerAuth(
        token_url="/login",
        token_secret="abc123",
        retrieve_user_handler=retrieve_user_handler,  # type: ignore[var-annotated]
    )

    @get("/login")
    def login_handler() -> Response["User"]:
        return jwt_auth.login(identifier=str(user.id))

    @get("/login_custom")
    def login_custom_handler() -> Response["User"]:
        return jwt_auth.login(identifier=str(user.id), response_body=user)

    with create_test_client(route_handlers=[login_custom_handler, login_handler]) as client:
        response = client.get("/login")
        response_custom = client.get("/login_custom")
        assert "access_token" in response.content.decode()
        assert response.content != response_custom.content

    assert jwt_auth.openapi_components.to_schema() == {
        "schemas": {},
        "securitySchemes": {
            "BearerToken": {
                "type": "oauth2",
                "description": "OAUTH2 password bearer authentication and authorization.",
                "name": "Authorization",
                "in": "header",
                "scheme": "Bearer",
                "bearerFormat": "JWT",
                "flows": {"password": {"tokenUrl": "/login"}},
            }
        },
    }

    assert jwt_auth.security_requirement == {"BearerToken": []}

    app = Litestar(on_app_init=[jwt_auth.on_app_init])
    assert app.openapi_schema.to_schema() == {
        "openapi": "3.1.0",
        "info": {"title": "Litestar API", "version": "1.0.0"},
        "servers": [{"url": "/"}],
        "paths": {},
        "components": {
            "schemas": {},
            "securitySchemes": {
                "BearerToken": {
                    "type": "oauth2",
                    "description": "OAUTH2 password bearer authentication and authorization.",
                    "name": "Authorization",
                    "in": "header",
                    "scheme": "Bearer",
                    "bearerFormat": "JWT",
                    "flows": {"password": {"tokenUrl": "/login"}},
                }
            },
        },
        "security": [{"BearerToken": []}],
    }


def test_type_encoders() -> None:
    # see: https://github.com/litestar-org/litestar/issues/1136

    class CustomUser:
        def __init__(self, id: str) -> None:
            self.id = id

    async def retrieve_user_handler(token: Token, connection: "ASGIConnection[Any, Any, Any, Any]") -> CustomUser:
        return CustomUser(id=token.sub)

    jwt_cookie_auth = JWTCookieAuth[User](
        retrieve_user_handler=retrieve_user_handler,
        token_secret="abc1234",
        exclude=["/"],
        type_encoders={CustomUser: lambda u: {"id": u.id}},
    )

    @get()
    def handler() -> Response[User]:
        data = CustomUser(id="1")
        return jwt_cookie_auth.login(identifier=str(data.id), response_body=data)

    with create_test_client([handler]) as client:
        response = client.get("/")
        assert response.status_code == HTTP_201_CREATED


async def retrieve_user_handler(token: Token, connection: "ASGIConnection[Any, Any, Any, Any]") -> Any:
    return User(name="moishe", id=uuid4())


@pytest.mark.parametrize(
    "config",
    (
        JWTAuth[User](
            retrieve_user_handler=retrieve_user_handler,
            token_secret="abc1234",
            exclude=["/"],
        ),
        JWTCookieAuth[User](
            retrieve_user_handler=retrieve_user_handler,
            token_secret="abc1234",
            exclude=["/"],
        ),
        OAuth2PasswordBearerAuth(
            token_url="/", exclude=["/"], token_secret="abc123", retrieve_user_handler=retrieve_user_handler
        ),
    ),
)
def test_returns_token_in_response_when_configured(config: JWTAuth) -> None:
    @get()
    def handler() -> Response[User]:
        return config.login(identifier="123", send_token_as_response_body=True)

    with create_test_client([handler]) as client:
        response = client.get("/")
        assert response.status_code == HTTP_201_CREATED
        assert isinstance(response.json(), dict) and response.json()


@pytest.mark.parametrize(
    "config",
    (
        JWTAuth[User](
            retrieve_user_handler=retrieve_user_handler,
            token_secret="abc1234",
            exclude=["/"],
        ),
        JWTCookieAuth[User](
            retrieve_user_handler=retrieve_user_handler,
            token_secret="abc1234",
            exclude=["/"],
        ),
        OAuth2PasswordBearerAuth(
            token_url="/", exclude=["/"], token_secret="abc123", retrieve_user_handler=retrieve_user_handler
        ),
    ),
)
def test_returns_none_when_response_body_is_none(config: JWTAuth) -> None:
    @get()
    def handler() -> Response[User]:
        return config.login(identifier="123", send_token_as_response_body=True, response_body=None)

    with create_test_client([handler]) as client:
        response = client.get("/")
        assert response.status_code == HTTP_201_CREATED
        assert response.json() is None



from typing import Any, Literal, Type, Union
from unittest import mock

import pytest

from litestar import Controller, Litestar, Router, get
from litestar.datastructures.url import URL
from litestar.enums import HttpMethod
from litestar.handlers.http_handlers.base import HTTPRouteHandler
from litestar.handlers.websocket_handlers.listener import (
    WebsocketListener,
    WebsocketListenerRouteHandler,
    websocket_listener,
)
from litestar.types.composite_types import TypeDecodersSequence

handler_decoder, router_decoder, controller_decoder, app_decoder = 4 * [(lambda t: t is URL, lambda t, v: URL(v))]


@pytest.fixture(scope="module")
def controller() -> Type[Controller]:
    class MyController(Controller):
        path = "/controller"
        type_decoders = [controller_decoder]

        @get("/http", type_decoders=[handler_decoder])
        def http(self) -> Any: ...

        @websocket_listener("/ws", type_decoders=[handler_decoder])
        async def handler(self, data: str) -> None: ...

    return MyController


@pytest.fixture(scope="module")
def websocket_listener_handler() -> Type[WebsocketListener]:
    class WebSocketHandler(WebsocketListener):
        path = "/ws-listener"
        type_decoders = [handler_decoder]

        def on_receive(self, data: str) -> None:  # pyright: ignore [reportIncompatibleMethodOverride]
            ...

    return WebSocketHandler


@pytest.fixture(scope="module")
def http_handler() -> HTTPRouteHandler:
    @get("/http", type_decoders=[handler_decoder])
    def http() -> Any: ...

    return http


@pytest.fixture(scope="module")
def websocket_handler() -> WebsocketListenerRouteHandler:
    @websocket_listener("/ws", type_decoders=[handler_decoder])
    async def websocket(data: str) -> None: ...

    return websocket


@pytest.fixture(scope="module")
def router(
    controller: Type[Controller],
    websocket_listener_handler: Type[WebsocketListenerRouteHandler],
    http_handler: Type[HTTPRouteHandler],
    websocket_handler: Type[WebsocketListenerRouteHandler],
) -> Router:
    return Router(
        "/router",
        type_decoders=[router_decoder],
        route_handlers=[controller, websocket_listener_handler, http_handler, websocket_handler],
    )


@pytest.fixture(scope="module")
@mock.patch("litestar.app.Litestar._get_default_plugins", mock.Mock(return_value=[]))
def app(router: Router) -> Litestar:
    return Litestar([router], type_decoders=[app_decoder])


@pytest.mark.parametrize(
    "path, method, type_decoders",
    (
        ("/router/controller/http", HttpMethod.GET, [app_decoder, router_decoder, controller_decoder, handler_decoder]),
        ("/router/controller/ws", "websocket", [app_decoder, router_decoder, controller_decoder, handler_decoder]),
        ("/router/http", HttpMethod.GET, [app_decoder, router_decoder, handler_decoder]),
        ("/router/ws", "websocket", [app_decoder, router_decoder, handler_decoder]),
        ("/router/ws-listener", "websocket", [app_decoder, router_decoder, handler_decoder]),
    ),
    ids=(
        "Controller http endpoint type decoders",
        "Controller ws endpoint type decoders",
        "Router http endpoint type decoders",
        "Router ws endpoint type decoders",
        "Router ws listener type decoders",
    ),
)
def test_resolve_type_decoders(
    path: str, method: Union[HttpMethod, Literal["websocket"]], type_decoders: TypeDecodersSequence, app: Litestar
) -> None:
    handler = app.route_handler_method_map[path][method]
    assert handler.resolve_type_decoders() == type_decoders



from typing import List

from litestar import get
from litestar.background_tasks import BackgroundTask, BackgroundTasks
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client


async def test_background_tasks_regular_execution() -> None:
    values: List[int] = []

    def extend_values(values_to_extend: List[int]) -> None:
        values.extend(values_to_extend)

    tasks = BackgroundTasks(
        [BackgroundTask(extend_values, [1, 2, 3]), BackgroundTask(extend_values, values_to_extend=[4, 5, 6])]
    )

    @get("/", background=tasks)
    def handler() -> None:
        return None

    with create_test_client(handler) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert values == [1, 2, 3, 4, 5, 6]


async def test_background_tasks_task_group_execution() -> None:
    values: List[int] = []

    def extend_values(values_to_extend: List[int]) -> None:
        values.extend(values_to_extend)

    tasks = BackgroundTasks(
        [BackgroundTask(extend_values, [1, 2, 3]), BackgroundTask(extend_values, values_to_extend=[4, 5, 6])],
        run_in_task_group=True,
    )

    @get("/", background=tasks)
    def handler() -> None:
        return None

    with create_test_client(handler) as client:
        response = client.get("/")
        assert response.status_code == HTTP_200_OK
        assert set(values) == {1, 2, 3, 4, 5, 6}



from attrs import define

from litestar import post
from litestar.status_codes import HTTP_201_CREATED
from litestar.testing import create_test_client


def test_parse_attrs_data_in_signature() -> None:
    @define(slots=True, frozen=True)
    class AttrsUser:
        name: str
        email: str

    @post("/")
    async def attrs_data(data: AttrsUser) -> AttrsUser:
        return data

    with create_test_client([attrs_data]) as client:
        response = client.post("/", json={"name": "foo", "email": "e@example.com"})
        assert response.status_code == HTTP_201_CREATED
        assert response.json().get("name") == "foo"
        assert response.json().get("email") == "e@example.com"



from __future__ import annotations

from typing import TYPE_CHECKING, Literal

if TYPE_CHECKING:
    import pydantic as pydantic_v2
    from pydantic import v1 as pydantic_v1
    from typing_extensions import TypeAlias

PydanticVersion = Literal["v1", "v2"]
BaseModelType: TypeAlias = "type[pydantic_v1.BaseModel| pydantic_v2.BaseModel]"



from __future__ import annotations

import warnings
from dataclasses import replace
from decimal import Decimal
from typing import Any, Generator, Generic, List, Optional, TypeVar

from msgspec import Meta
from typing_extensions import Annotated

from litestar.dto import AbstractDTO, DTOField, Mark
from litestar.dto.data_structures import DTOFieldDefinition
from litestar.exceptions import LitestarWarning, MissingDependencyException
from litestar.types import Empty
from litestar.typing import FieldDefinition
from litestar.utils import warn_deprecation

try:
    from piccolo.columns import Column, column_types
    from piccolo.table import Table
except ImportError as e:
    raise MissingDependencyException("piccolo") from e


T = TypeVar("T", bound=Table)

__all__ = ("PiccoloDTO",)


def __getattr__(name: str) -> Any:
    warn_deprecation(
        deprecated_name=f"litestar.contrib.piccolo.{name}",
        version="2.3.2",
        kind="import",
        removal_in="3.0.0",
        info="importing from 'litestar.contrib.piccolo' is deprecated and will be removed in 3.0, please import from 'litestar_piccolo' package directly instead",
    )
    return getattr(name, name)


def _parse_piccolo_type(column: Column, extra: dict[str, Any]) -> FieldDefinition:
    is_optional = not column._meta.required

    if isinstance(column, (column_types.Decimal, column_types.Numeric)):
        column_type: Any = Decimal
        meta = Meta(extra=extra)
    elif isinstance(column, (column_types.Email, column_types.Varchar)):
        column_type = str
        if is_optional:
            meta = Meta(extra=extra)
            warnings.warn(
                f"Dropping max_length constraint for column {column!r} because the " "column is optional",
                category=LitestarWarning,
                stacklevel=2,
            )
        else:
            meta = Meta(max_length=column.length, extra=extra)
    elif isinstance(column, column_types.Array):
        column_type = List[column.base_column.value_type]  # type: ignore[name-defined]
        meta = Meta(extra=extra)
    elif isinstance(column, (column_types.JSON, column_types.JSONB)):
        column_type = str
        meta = Meta(extra={**extra, "format": "json"})
    elif isinstance(column, column_types.Text):
        column_type = str
        meta = Meta(extra={**extra, "format": "text-area"})
    else:
        column_type = column.value_type
        meta = Meta(extra=extra)

    if is_optional:
        column_type = Optional[column_type]

    return FieldDefinition.from_annotation(Annotated[column_type, meta])


def _create_column_extra(column: Column) -> dict[str, Any]:
    extra: dict[str, Any] = {}

    if column._meta.help_text:
        extra["description"] = column._meta.help_text

    if column._meta.get_choices_dict():
        extra["enum"] = column._meta.get_choices_dict()

    return extra


class PiccoloDTO(AbstractDTO[T], Generic[T]):
    @classmethod
    def generate_field_definitions(cls, model_type: type[Table]) -> Generator[DTOFieldDefinition, None, None]:
        for column in model_type._meta.columns:
            mark = Mark.WRITE_ONLY if column._meta.secret else Mark.READ_ONLY if column._meta.primary_key else None
            yield replace(
                DTOFieldDefinition.from_field_definition(
                    field_definition=_parse_piccolo_type(column, _create_column_extra(column)),
                    dto_field=DTOField(mark=mark),
                    model_name=model_type.__name__,
                    default_factory=None,
                ),
                default=Empty if column._meta.required else None,
                name=column._meta.name,
            )

    @classmethod
    def detect_nested_field(cls, field_definition: FieldDefinition) -> bool:
        return field_definition.is_subclass_of(Table)



from typing import TYPE_CHECKING, Any, Tuple

from litestar import Controller, HttpMethod, Litestar, Response, Router, get
from litestar.testing import create_test_client

if TYPE_CHECKING:
    from litestar.types import Serializer


def create_mock_encoder(name: str) -> Tuple[type, "Serializer"]:
    mock_type = type(name, (type,), {})

    def mock_encoder(obj: Any) -> Any:
        return name

    return mock_type, mock_encoder


handler_type, handler_encoder = create_mock_encoder("HandlerType")
router_type, router_encoder = create_mock_encoder("RouterType")
controller_type, controller_encoder = create_mock_encoder("ControllerType")
app_type, app_encoder = create_mock_encoder("AppType")


def test_resolve_type_encoders() -> None:
    class MyController(Controller):
        type_encoders = {controller_type: controller_encoder}

        @get("/", type_encoders={handler_type: handler_encoder})
        def handler(self) -> Any: ...

    router = Router("/router", type_encoders={router_type: router_encoder}, route_handlers=[MyController])
    app = Litestar([router], type_encoders={app_type: app_encoder})

    route_handler = app.routes[0].route_handler_map[HttpMethod.GET][0]  # type: ignore[union-attr]
    encoders = route_handler.resolve_type_encoders()
    assert encoders.get(handler_type) == handler_encoder
    assert encoders.get(controller_type) == controller_encoder
    assert encoders.get(router_type) == router_encoder
    assert encoders.get(app_type) == app_encoder


def test_type_encoders_response_override() -> None:
    class Foo:
        pass

    @get("/", type_encoders={Foo: lambda f: "foo"})
    def handler() -> Response:
        return Response({"obj": Foo()}, type_encoders={Foo: lambda f: "FOO"})

    with create_test_client([handler]) as client:
        assert client.get("/").json() == {"obj": "FOO"}



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Literal, NamedTuple

from litestar.utils.deprecation import warn_deprecation

__all__ = (
    "ControllerRouterHandler",
    "PathParameterDefinition",
    "PathParameterDefinition",
    "ReservedKwargs",
    "RouteHandlerMapItem",
    "RouteHandlerType",
)

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    from litestar.app import Litestar
    from litestar.controller import Controller
    from litestar.handlers.asgi_handlers import ASGIRouteHandler
    from litestar.handlers.http_handlers import HTTPRouteHandler
    from litestar.handlers.websocket_handlers import WebsocketRouteHandler
    from litestar.router import Router
    from litestar.template import TemplateConfig
    from litestar.template.config import EngineType
    from litestar.types import Method

ReservedKwargs: TypeAlias = Literal["request", "socket", "headers", "query", "cookies", "state", "data"]
RouteHandlerType: TypeAlias = "HTTPRouteHandler | WebsocketRouteHandler | ASGIRouteHandler"
ControllerRouterHandler: TypeAlias = "type[Controller] | RouteHandlerType | Router | Callable[..., Any]"
RouteHandlerMapItem: TypeAlias = 'dict[Method | Literal["websocket", "asgi"], RouteHandlerType]'
TemplateConfigType: TypeAlias = "TemplateConfig[EngineType]"

# deprecated
_LitestarType: TypeAlias = "Litestar"


class PathParameterDefinition(NamedTuple):
    """Path parameter tuple."""

    name: str
    full: str
    type: type
    parser: Callable[[str], Any] | None


def __getattr__(name: str) -> Any:
    if name == "LitestarType":
        warn_deprecation(
            "2.2.1",
            "LitestarType",
            "import",
            removal_in="3.0.0",
            alternative="Litestar",
        )
        return _LitestarType
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



from datetime import datetime, timedelta

from time_machine import travel

from litestar.datastructures import Cookie


def test_basic_cookie_as_header() -> None:
    cookie = Cookie(key="key")
    assert cookie.to_header() == 'Set-Cookie: key=""; Path=/; SameSite=lax'


@travel(datetime.utcnow, tick=False)
def test_cookie_as_header() -> None:
    expires_sec = 123
    cookie = Cookie(
        key="key",
        value="value",
        path="/path",
        expires=expires_sec,
        domain="domain.com",
        secure=True,
        httponly=True,
        samesite="strict",
    )

    now = datetime.utcnow()
    expected_expired = (now + timedelta(seconds=expires_sec)).strftime("%a, %d %b %Y %H:%M:%S GMT")
    assert cookie.to_header() == (
        f"Set-Cookie: key=value; Domain=domain.com; expires={expected_expired}; HttpOnly; Path=/path; SameSite=strict; Secure"
    )


def test_cookie_with_max_age_as_header() -> None:
    cookie = Cookie(key="key", max_age=10)
    assert cookie.to_header() == 'Set-Cookie: key=""; Max-Age=10; Path=/; SameSite=lax'


def test_cookie_as_header_without_header_name() -> None:
    cookie = Cookie(key="key")
    assert cookie.to_header(header="") == 'key=""; Path=/; SameSite=lax'


def test_equality() -> None:
    assert Cookie(key="key") == Cookie(key="key")
    assert Cookie(key="key") != Cookie(key="key", path="/test")
    assert Cookie(key="key", path="/test") != Cookie(key="key", path="/test", domain="localhost")
    assert Cookie(key="key", path="/test", domain="localhost") == Cookie(key="key", path="/test", domain="localhost")
    assert Cookie(key="key") != "key"



from typing import List, Optional, Type

import pytest

from litestar import Controller
from litestar.enums import MediaType
from litestar.openapi.config import OpenAPIConfig
from litestar.openapi.controller import OpenAPIController
from litestar.openapi.plugins import (
    JsonRenderPlugin,
    OpenAPIRenderPlugin,
    RapidocRenderPlugin,
    RedocRenderPlugin,
    ScalarRenderPlugin,
    StoplightRenderPlugin,
    SwaggerRenderPlugin,
)
from litestar.status_codes import HTTP_200_OK, HTTP_404_NOT_FOUND
from litestar.testing import create_test_client

root_paths: List[str] = ["", "/part1", "/part1/part2"]


@pytest.fixture()
def config(openapi_controller: Optional[Type[OpenAPIController]]) -> OpenAPIConfig:
    return OpenAPIConfig(title="Litestar API", version="1.0.0", openapi_controller=openapi_controller)


def test_default_redoc_cdn_urls(
    person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    default_redoc_version = "next"
    default_redoc_js_bundle = f"https://cdn.jsdelivr.net/npm/redoc@{default_redoc_version}/bundles/redoc.standalone.js"
    with create_test_client([person_controller, pet_controller], openapi_config=config) as client:
        response = client.get("/schema/redoc")
        assert default_redoc_js_bundle in response.text


def test_default_swagger_ui_cdn_urls(
    person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    default_swagger_ui_version = "5.1.3"
    default_swagger_bundles = [
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{default_swagger_ui_version}/swagger-ui.css",
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{default_swagger_ui_version}/swagger-ui-bundle.js",
        f"https://cdn.jsdelivr.net/npm/swagger-ui-dist@{default_swagger_ui_version}/swagger-ui-standalone-preset.js",
    ]
    with create_test_client([person_controller, pet_controller], openapi_config=config) as client:
        response = client.get("/schema/swagger")
        assert all(cdn_url in response.text for cdn_url in default_swagger_bundles)


def test_default_stoplight_elements_cdn_urls(
    person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    default_stoplight_elements_version = "7.7.18"
    default_stoplight_elements_bundles = [
        f"https://unpkg.com/@stoplight/elements@{default_stoplight_elements_version}/styles.min.css",
        f"https://unpkg.com/@stoplight/elements@{default_stoplight_elements_version}/web-components.min.js",
    ]
    with create_test_client([person_controller, pet_controller], openapi_config=config) as client:
        response = client.get("/schema/elements")
        assert all(cdn_url in response.text for cdn_url in default_stoplight_elements_bundles)


def test_default_rapidoc_elements_cdn_urls(
    person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    default_rapidoc_version = "9.3.4"
    default_rapidoc_bundles = [f"https://unpkg.com/rapidoc@{default_rapidoc_version}/dist/rapidoc-min.js"]
    with create_test_client([person_controller, pet_controller], openapi_config=config) as client:
        response = client.get("/schema/rapidoc")
        assert all(cdn_url in response.text for cdn_url in default_rapidoc_bundles)


def test_redoc_with_google_fonts(
    person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    google_font_cdn = "https://fonts.googleapis.com/css?family=Montserrat:300,400,700|Roboto:300,400,700"
    with create_test_client([person_controller, pet_controller], openapi_config=config) as client:
        response = client.get("/schema/redoc")
        assert google_font_cdn in response.text


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (type("OfflineOpenAPIController", (OpenAPIController,), {"redoc_google_fonts": False}), []),
        (None, [RedocRenderPlugin(google_fonts=False)]),
    ],
)
def test_redoc_without_google_fonts(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/redoc")
        assert "fonts.googleapis.com" not in response.text


OFFLINE_LOCATION_JS_URL = "https://offline_location/bundle.js"
OFFLINE_LOCATION_CSS_URL = "https://offline_location/bundle.css"
OFFLINE_LOCATION_OTHER_URL = "https://offline_location/bundle.other"


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (type("OfflineOpenAPIController", (OpenAPIController,), {"redoc_js_url": OFFLINE_LOCATION_JS_URL}), []),
        (None, [RedocRenderPlugin(js_url=OFFLINE_LOCATION_JS_URL)]),
    ],
)
def test_openapi_redoc_offline(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/redoc")
        assert OFFLINE_LOCATION_JS_URL in response.text


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (
            type(
                "OfflineOpenAPIController",
                (OpenAPIController,),
                {
                    "swagger_ui_bundle_js_url": OFFLINE_LOCATION_JS_URL,
                    "swagger_css_url": OFFLINE_LOCATION_CSS_URL,
                    "swagger_ui_standalone_preset_js_url": OFFLINE_LOCATION_OTHER_URL,
                },
            ),
            [],
        ),
        (
            None,
            [
                SwaggerRenderPlugin(
                    js_url=OFFLINE_LOCATION_JS_URL,
                    css_url=OFFLINE_LOCATION_CSS_URL,
                    standalone_preset_js_url=OFFLINE_LOCATION_OTHER_URL,
                )
            ],
        ),
    ],
)
def test_openapi_swagger_offline(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/swagger")
        assert all(
            offline_url in response.text
            for offline_url in [OFFLINE_LOCATION_JS_URL, OFFLINE_LOCATION_CSS_URL, OFFLINE_LOCATION_OTHER_URL]
        )


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (
            type(
                "OfflineOpenAPIController",
                (OpenAPIController,),
                {
                    "stoplight_elements_css_url": OFFLINE_LOCATION_CSS_URL,
                    "stoplight_elements_js_url": OFFLINE_LOCATION_JS_URL,
                },
            ),
            [],
        ),
        (
            None,
            [
                StoplightRenderPlugin(
                    js_url=OFFLINE_LOCATION_JS_URL,
                    css_url=OFFLINE_LOCATION_CSS_URL,
                )
            ],
        ),
    ],
)
def test_openapi_stoplight_elements_offline(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/elements")
        assert all(offline_url in response.text for offline_url in [OFFLINE_LOCATION_JS_URL, OFFLINE_LOCATION_CSS_URL])


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (
            None,
            [
                ScalarRenderPlugin(
                    js_url=OFFLINE_LOCATION_JS_URL,
                    css_url=OFFLINE_LOCATION_CSS_URL,
                )
            ],
        ),
    ],
)
def test_openapi_scalar_offline(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/scalar")
        assert all(offline_url in response.text for offline_url in [OFFLINE_LOCATION_JS_URL, OFFLINE_LOCATION_CSS_URL])


@pytest.mark.parametrize(
    ("openapi_controller", "render_plugins"),
    [
        (type("OfflineOpenAPIController", (OpenAPIController,), {"rapidoc_js_url": OFFLINE_LOCATION_JS_URL}), []),
        (None, [RapidocRenderPlugin(js_url=OFFLINE_LOCATION_JS_URL)]),
    ],
)
def test_openapi_rapidoc_offline(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
    render_plugins: List[OpenAPIRenderPlugin],
) -> None:
    offline_config = OpenAPIConfig(
        title="Litestar API", version="1.0.0", openapi_controller=openapi_controller, render_plugins=render_plugins
    )
    with create_test_client([person_controller, pet_controller], openapi_config=offline_config) as client:
        response = client.get("/schema/rapidoc")
        assert OFFLINE_LOCATION_JS_URL in response.text


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_root(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        response = client.get("/schema")
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_redoc(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        response = client.get("/schema/redoc")
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_swagger(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        response = client.get("/schema/swagger")
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_swagger_caching_schema(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        # Make sure that the schema is tweaked for swagger as the openapi version is changed.
        # Because schema can get cached, make sure that getting a different schema type before works.
        client.get("/schema/redoc")  # Cache the schema
        response = client.get("/schema/swagger")  # Request swagger, should use a different cache

        assert "3.1.0" in response.text  # Make sure the injected version is still there
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_stoplight_elements(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        response = client.get("/schema/elements/")
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize("root_path", root_paths)
def test_openapi_rapidoc(
    root_path: str, person_controller: Type[Controller], pet_controller: Type[Controller], config: OpenAPIConfig
) -> None:
    with create_test_client([person_controller, pet_controller], root_path=root_path, openapi_config=config) as client:
        response = client.get("/schema/rapidoc")
        assert response.status_code == HTTP_200_OK
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


def test_openapi_root_not_allowed(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
) -> None:
    with create_test_client(
        [person_controller, pet_controller],
        openapi_config=OpenAPIConfig(
            title="Litestar API",
            version="1.0.0",
            enabled_endpoints={"swagger", "elements", "openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=openapi_controller,
        ),
    ) as client:
        response = client.get("/schema")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


def test_openapi_redoc_not_allowed(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
) -> None:
    with create_test_client(
        [person_controller, pet_controller],
        openapi_config=OpenAPIConfig(
            title="Litestar API",
            version="1.0.0",
            enabled_endpoints={"swagger", "elements", "openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=openapi_controller,
        ),
    ) as client:
        response = client.get("/schema/redoc")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


def test_openapi_swagger_not_allowed(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
) -> None:
    with create_test_client(
        [person_controller, pet_controller],
        openapi_config=OpenAPIConfig(
            title="Litestar API",
            version="1.0.0",
            enabled_endpoints={"redoc", "elements", "openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=openapi_controller,
        ),
    ) as client:
        response = client.get("/schema/swagger")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


def test_openapi_stoplight_elements_not_allowed(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
) -> None:
    with create_test_client(
        [person_controller, pet_controller],
        openapi_config=OpenAPIConfig(
            title="Litestar API",
            version="1.0.0",
            enabled_endpoints={"redoc", "swagger", "openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=openapi_controller,
        ),
    ) as client:
        response = client.get("/schema/elements/")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


def test_openapi_rapidoc_not_allowed(
    person_controller: Type[Controller],
    pet_controller: Type[Controller],
    openapi_controller: Optional[Type[OpenAPIController]],
) -> None:
    with create_test_client(
        [person_controller, pet_controller],
        openapi_config=OpenAPIConfig(
            title="Litestar API",
            version="1.0.0",
            enabled_endpoints={"swagger", "elements", "openapi.json", "openapi.yaml", "openapi.yml"},
            openapi_controller=openapi_controller,
        ),
    ) as client:
        response = client.get("/schema/rapidoc")
        assert response.status_code == HTTP_404_NOT_FOUND
        assert response.headers["content-type"].startswith(MediaType.HTML.value)


@pytest.mark.parametrize(
    ("render_plugins",),
    [
        ([],),
        ([RedocRenderPlugin()],),
        ([RedocRenderPlugin(), JsonRenderPlugin()],),
        ([JsonRenderPlugin(path="/custom_path")],),
        ([JsonRenderPlugin(path=["/openapi.json", "/custom_path"])],),
    ],
)
def test_json_plugin_always_enabled(render_plugins: List["OpenAPIRenderPlugin"]) -> None:
    """We assume that an '/openapi.json' path is available in many of the openapi render plugins.

    This test ensures that the json plugin is always enabled, even if the user has not explicitly
    included it in the render_plugins list.
    """

    openapi_config = OpenAPIConfig(title="my title", version="1.0.0", render_plugins=render_plugins)
    with create_test_client([], openapi_config=openapi_config) as client:
        response = client.get("/schema/openapi.json")
        assert response.status_code == HTTP_200_OK


def test_default_plugin_explicit_path() -> None:
    config = OpenAPIConfig(title="my title", version="1.0.0", render_plugins=[SwaggerRenderPlugin(path="/")])
    with create_test_client([], openapi_config=config) as client:
        response = client.get("/schema/")
        assert response.status_code == HTTP_200_OK

        response = client.get("/schema/swagger")
        assert response.status_code == HTTP_404_NOT_FOUND


def test_default_plugin_backward_compatibility() -> None:
    config = OpenAPIConfig(title="my title", version="1.0.0")
    with create_test_client([], openapi_config=config) as client:
        response = client.get("/schema/")
        assert response.status_code == HTTP_200_OK

        response = client.get("/schema/redoc")
        assert response.status_code == HTTP_200_OK


def test_default_plugin_backward_compatibility_not_found() -> None:
    config = OpenAPIConfig(title="my title", version="1.0.0", enabled_endpoints={"redoc"}, root_schema_site="swagger")
    with create_test_client([], openapi_config=config) as client:
        response = client.get("/schema/")
        assert response.status_code == HTTP_404_NOT_FOUND

        response = client.get("/schema/swagger")
        assert response.status_code == HTTP_404_NOT_FOUND

        response = client.get("/schema/redoc")
        assert response.status_code == HTTP_200_OK


def test_default_plugin_future_compatibility() -> None:
    config = OpenAPIConfig(title="my title", version="1.0.0", render_plugins=[SwaggerRenderPlugin()])
    with create_test_client([], openapi_config=config) as client:
        response = client.get("/schema/")
        assert response.status_code == HTTP_200_OK

        response = client.get("/schema/swagger")
        assert response.status_code == HTTP_200_OK



from typing import Optional, Type

import pytest

from litestar import get
from litestar.params import Parameter, ParameterKwarg
from litestar.status_codes import HTTP_200_OK, HTTP_400_BAD_REQUEST
from litestar.testing import create_test_client


@pytest.mark.parametrize(
    "t_type,param_dict,param,expected_code",
    [
        (
            Optional[str],
            {},
            Parameter(cookie="special-cookie", min_length=1, max_length=2, required=False),
            HTTP_200_OK,
        ),
        (int, {"special-cookie": "123"}, Parameter(cookie="special-cookie", ge=100, le=201), HTTP_200_OK),
        (int, {"special-cookie": "123"}, Parameter(cookie="special-cookie", ge=100, le=120), HTTP_400_BAD_REQUEST),
        (int, {}, Parameter(cookie="special-cookie", ge=100, le=120), HTTP_400_BAD_REQUEST),
        (Optional[int], {}, Parameter(cookie="special-cookie", ge=100, le=120, required=False), HTTP_200_OK),
    ],
)
def test_cookie_params(t_type: Type, param_dict: dict, param: ParameterKwarg, expected_code: int) -> None:
    test_path = "/test"

    @get(path=test_path)
    def test_method(special_cookie: t_type = param) -> None:  # type: ignore[valid-type]
        if special_cookie:
            assert special_cookie in (param_dict.get("special-cookie"), int(param_dict.get("special-cookie")))  # type: ignore[arg-type]

    with create_test_client(test_method) as client:
        # Set cookies on the client to avoid warnings about per-request cookies.
        client.cookies = param_dict  # type: ignore[assignment]
        response = client.get(test_path)
        assert response.status_code == expected_code, response.json()



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, cast

import pytest
from pydantic import v1 as pydantic_v1

from litestar import Request, post
from litestar.contrib.pydantic import PydanticDTO, _model_dump_json
from litestar.dto import DTOConfig
from litestar.testing import create_test_client
from litestar.types import Empty
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import ModuleType

    from pydantic import BaseModel

    from litestar import Litestar


def test_schema_required_fields_with_pydantic_dto(
    use_experimental_dto_backend: bool, base_model: type[BaseModel]
) -> None:
    class PydanticUser(base_model):  # type: ignore[misc, valid-type]
        age: int
        name: str

    class UserDTO(PydanticDTO[PydanticUser]):
        config = DTOConfig(experimental_codegen_backend=use_experimental_dto_backend)

    @post(dto=UserDTO, return_dto=None, signature_types=[PydanticUser])
    def handler(data: PydanticUser, request: Request) -> dict:
        schema = request.app.openapi_schema
        return schema.to_schema()

    with create_test_client(handler) as client:
        data = PydanticUser(name="A", age=10)
        headers = {"Content-Type": "application/json; charset=utf-8"}
        received = client.post(
            "/",
            content=_model_dump_json(data),
            headers=headers,
        )
        required = next(iter(received.json()["components"]["schemas"].values()))["required"]
        assert len(required) == 2


def test_field_definition_implicit_optional_default(base_model: type[BaseModel]) -> None:
    class Model(base_model):  # type: ignore[misc, valid-type]
        a: Optional[str]  # noqa: UP007

    dto_type = PydanticDTO[Model]
    field_defs = list(dto_type.generate_field_definitions(Model))
    assert len(field_defs) == 1
    assert field_defs[0].default is None


def test_detect_nested_field_pydantic_v1(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("litestar.contrib.pydantic.pydantic_dto_factory.pydantic_v2", Empty)

    class Model(pydantic_v1.BaseModel):
        a: str

    dto_type = PydanticDTO[Model]
    assert dto_type.detect_nested_field(FieldDefinition.from_annotation(Model)) is True
    assert dto_type.detect_nested_field(FieldDefinition.from_annotation(int)) is False


def test_pydantic_field_descriptions(create_module: Callable[[str], ModuleType]) -> None:
    module = create_module(
        """
from litestar import Litestar, get
from litestar.contrib.pydantic import PydanticDTO
from litestar.dto import DTOConfig
from pydantic import BaseModel, Field
from typing_extensions import Annotated

class User(BaseModel):
    id: Annotated[
        int,
        Field(description="This is a test (id description)."),
    ]

class DataCollectionDTO(PydanticDTO[User]):
    config = DTOConfig(rename_strategy="camel")

@get("/user", return_dto=DataCollectionDTO, sync_to_thread=False)
def get_user() -> User:
    return User(id=user_id)

app = Litestar(route_handlers=[get_user])
        """
    )
    app = cast("Litestar", module.app)
    schema = app.openapi_schema
    assert schema.components.schemas is not None
    component_schema = schema.components.schemas["GetUserUserResponseBody"]
    assert component_schema.properties is not None
    assert component_schema.properties["id"].description == "This is a test (id description)."



from __future__ import annotations

import importlib.util
import logging
import os
import random
import string
import sys
from datetime import datetime
from os import urandom
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncGenerator, Callable, Generator, TypeVar, Union, cast
from unittest.mock import AsyncMock, MagicMock

import pytest
from pytest_lazy_fixtures import lf
from redis.asyncio import Redis as AsyncRedis
from redis.client import Redis
from time_machine import travel

from litestar.logging import LoggingConfig
from litestar.logging.config import default_handlers as logging_default_handlers
from litestar.middleware.session import SessionMiddleware
from litestar.middleware.session.base import BaseSessionBackend
from litestar.middleware.session.client_side import ClientSideSessionBackend, CookieBackendConfig
from litestar.middleware.session.server_side import ServerSideSessionBackend, ServerSideSessionConfig
from litestar.openapi.config import OpenAPIConfig
from litestar.stores.base import Store
from litestar.stores.file import FileStore
from litestar.stores.memory import MemoryStore
from litestar.stores.redis import RedisStore
from litestar.testing import RequestFactory

if TYPE_CHECKING:
    from types import ModuleType

    from pytest import FixtureRequest, MonkeyPatch
    from time_machine import Coordinates

    from litestar import Litestar
    from litestar.types import (
        AnyIOBackend,
        ASGIApp,
        ASGIVersion,
        GetLogger,
        Receive,
        RouteHandlerType,
        Scope,
        ScopeSession,
        Send,
    )


pytest_plugins = ["tests.docker_service_fixtures"]


@pytest.fixture
def mock() -> MagicMock:
    return MagicMock()


@pytest.fixture()
def async_mock() -> AsyncMock:
    return AsyncMock()


@pytest.fixture(params=[pytest.param("asyncio", id="asyncio"), pytest.param("trio", id="trio")])
def anyio_backend(request: pytest.FixtureRequest) -> str:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture()
def mock_asgi_app() -> ASGIApp:
    async def asgi_app(scope: Scope, receive: Receive, send: Send) -> None: ...

    return asgi_app


@pytest.fixture()
def redis_store(redis_client: AsyncRedis) -> RedisStore:
    return RedisStore(redis=redis_client)


@pytest.fixture()
def memory_store() -> MemoryStore:
    return MemoryStore()


@pytest.fixture()
def file_store(tmp_path: Path) -> FileStore:
    return FileStore(path=tmp_path)


@pytest.fixture(
    params=[pytest.param("redis_store", marks=pytest.mark.xdist_group("redis")), "memory_store", "file_store"]
)
def store(request: FixtureRequest) -> Store:
    return cast("Store", request.getfixturevalue(request.param))


@pytest.fixture
def cookie_session_backend_config() -> CookieBackendConfig:
    return CookieBackendConfig(secret=urandom(16))


@pytest.fixture()
def cookie_session_backend(cookie_session_backend_config: CookieBackendConfig) -> ClientSideSessionBackend:
    return ClientSideSessionBackend(config=cookie_session_backend_config)


@pytest.fixture(
    params=[
        pytest.param(lf("cookie_session_backend_config"), id="cookie"),
        pytest.param(lf("server_side_session_config"), id="server-side"),
    ]
)
def session_backend_config(request: pytest.FixtureRequest) -> ServerSideSessionConfig | CookieBackendConfig:
    return cast("Union[ServerSideSessionConfig, CookieBackendConfig]", request.param)


@pytest.fixture()
def server_side_session_config() -> ServerSideSessionConfig:
    return ServerSideSessionConfig()


@pytest.fixture()
def server_side_session_backend(server_side_session_config: ServerSideSessionConfig) -> ServerSideSessionBackend:
    return ServerSideSessionBackend(config=server_side_session_config)


@pytest.fixture(
    params=[
        pytest.param("cookie_session_backend", id="cookie"),
        pytest.param("server_side_session_backend", id="server-side"),
    ]
)
def session_backend(request: pytest.FixtureRequest) -> BaseSessionBackend:
    return cast("BaseSessionBackend", request.getfixturevalue(request.param))


@pytest.fixture()
def session_backend_config_memory(memory_store: MemoryStore) -> ServerSideSessionConfig:
    return ServerSideSessionConfig()


@pytest.fixture
def session_middleware(session_backend: BaseSessionBackend, mock_asgi_app: ASGIApp) -> SessionMiddleware[Any]:
    return SessionMiddleware(app=mock_asgi_app, backend=session_backend)


@pytest.fixture
def cookie_session_middleware(
    cookie_session_backend: ClientSideSessionBackend, mock_asgi_app: ASGIApp
) -> SessionMiddleware[ClientSideSessionBackend]:
    return SessionMiddleware(app=mock_asgi_app, backend=cookie_session_backend)


@pytest.fixture
def test_client_backend(anyio_backend_name: str) -> AnyIOBackend:
    return cast("AnyIOBackend", anyio_backend_name)


@pytest.fixture
def create_scope() -> Callable[..., Scope]:
    def inner(
        *,
        type: str = "http",
        app: Litestar | None = None,
        asgi: ASGIVersion | None = None,
        auth: Any = None,
        client: tuple[str, int] | None = ("testclient", 50000),
        extensions: dict[str, dict[object, object]] | None = None,
        http_version: str = "1.1",
        path: str = "/",
        path_params: dict[str, str] | None = None,
        query_string: str = "",
        root_path: str = "",
        route_handler: RouteHandlerType | None = None,
        scheme: str = "http",
        server: tuple[str, int | None] | None = ("testserver", 80),
        session: ScopeSession | None = None,
        state: dict[str, Any] | None = None,
        user: Any = None,
        **kwargs: dict[str, Any],
    ) -> Scope:
        scope = {
            "app": app,
            "asgi": asgi or {"spec_version": "2.0", "version": "3.0"},
            "auth": auth,
            "type": type,
            "path": path,
            "raw_path": path.encode(),
            "root_path": root_path,
            "scheme": scheme,
            "query_string": query_string.encode(),
            "client": client,
            "server": server,
            "method": "GET",
            "http_version": http_version,
            "extensions": extensions or {"http.response.template": {}},
            "state": state or {},
            "path_params": path_params or {},
            "route_handler": route_handler,
            "user": user,
            "session": session,
            **kwargs,
        }
        return cast("Scope", scope)

    return inner


@pytest.fixture
def scope(create_scope: Callable[..., Scope]) -> Scope:
    return create_scope()


@pytest.fixture
def create_module(tmp_path: Path, monkeypatch: MonkeyPatch) -> Callable[[str], ModuleType]:
    """Utility fixture for dynamic module creation."""

    def wrapped(source: str) -> ModuleType:
        """

        Args:
            source: Source code as a string.

        Returns:
            An imported module.
        """
        T = TypeVar("T")

        def not_none(val: T | T | None) -> T:
            assert val is not None
            return val

        def module_name_generator() -> str:
            letters = string.ascii_lowercase
            return "".join(random.choice(letters) for _ in range(10))

        module_name = module_name_generator()
        path = tmp_path / f"{module_name}.py"
        path.write_text(source)
        # https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
        spec = not_none(importlib.util.spec_from_file_location(module_name, path))
        module = not_none(importlib.util.module_from_spec(spec))
        monkeypatch.setitem(sys.modules, module_name, module)
        not_none(spec.loader).exec_module(module)
        return module

    return wrapped


@pytest.fixture()
def frozen_datetime() -> Generator[Coordinates, None, None]:
    with travel(datetime.utcnow, tick=False) as frozen:
        yield frozen


@pytest.fixture()
def request_factory() -> RequestFactory:
    return RequestFactory()


@pytest.fixture()
def reset_httpx_logging() -> Generator[None, None, None]:
    # ensure that httpx logging is not interfering with our test client
    httpx_logger = logging.getLogger("httpx")
    initial_level = httpx_logger.level
    httpx_logger.setLevel(logging.WARNING)
    yield
    httpx_logger.setLevel(initial_level)


# the monkeypatch fixture does not work with session scoped dependencies
@pytest.fixture(autouse=True, scope="session")
def disable_warn_implicit_sync_to_thread() -> Generator[None, None, None]:
    old_value = os.getenv("LITESTAR_WARN_IMPLICIT_SYNC_TO_THREAD")
    os.environ["LITESTAR_WARN_IMPLICIT_SYNC_TO_THREAD"] = "0"
    yield
    if old_value is not None:
        os.environ["LITESTAR_WARN_IMPLICIT_SYNC_TO_THREAD"] = old_value


@pytest.fixture()
def disable_warn_sync_to_thread_with_async(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("LITESTAR_WARN_SYNC_TO_THREAD_WITH_ASYNC", "0")


@pytest.fixture()
def enable_warn_implicit_sync_to_thread(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("LITESTAR_WARN_IMPLICIT_SYNC_TO_THREAD", "1")


@pytest.fixture
def get_logger() -> GetLogger:
    # due to the limitations of caplog we have to place this call here.
    # we also have to allow propagation.
    return LoggingConfig(
        handlers=logging_default_handlers,
        loggers={
            "litestar": {"level": "INFO", "handlers": ["queue_listener"], "propagate": True},
        },
    ).configure()


@pytest.fixture()
async def redis_client(docker_ip: str, redis_service: None) -> AsyncGenerator[AsyncRedis, None]:
    # this is to get around some weirdness with pytest-asyncio and redis interaction
    # on 3.8 and 3.9

    Redis(host=docker_ip, port=6397).flushall()
    client: AsyncRedis = AsyncRedis(host=docker_ip, port=6397)
    yield client
    try:
        await client.aclose()  # type: ignore[attr-defined]
    except RuntimeError:
        pass


@pytest.fixture(autouse=True)
def _patch_openapi_config(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("litestar.app.DEFAULT_OPENAPI_CONFIG", OpenAPIConfig(title="Litestar API", version="1.0.0"))



from litestar.app import Litestar
from litestar.connection import Request, WebSocket
from litestar.controller import Controller
from litestar.enums import HttpMethod, MediaType
from litestar.handlers import asgi, delete, get, head, patch, post, put, route, websocket, websocket_listener
from litestar.response import Response
from litestar.router import Router
from litestar.utils.version import get_version

__version__ = get_version()


__all__ = (
    "Controller",
    "HttpMethod",
    "Litestar",
    "MediaType",
    "Request",
    "Response",
    "Router",
    "WebSocket",
    "__version__",
    "asgi",
    "delete",
    "get",
    "head",
    "patch",
    "post",
    "put",
    "route",
    "websocket",
    "websocket_listener",
)



from typing import Type, Union

import pytest

from litestar import Controller, Litestar, Router, WebSocket
from litestar.handlers.websocket_handlers.listener import WebsocketListener, websocket_listener

RouterWebSocket: Type[WebSocket] = type("RouterWebSocket", (WebSocket,), {})
ControllerWebSocket: Type[WebSocket] = type("ControllerWebSocket", (WebSocket,), {})
AppWebSocket: Type[WebSocket] = type("AppWebSocket", (WebSocket,), {})
HandlerWebSocket: Type[WebSocket] = type("HandlerWebSocket", (WebSocket,), {})


@pytest.mark.parametrize(
    "handler_websocket_class, controller_websocket_class, router_websocket_class, app_websocket_class, has_default_app_class, expected",
    (
        (HandlerWebSocket, ControllerWebSocket, RouterWebSocket, AppWebSocket, True, HandlerWebSocket),
        (None, ControllerWebSocket, RouterWebSocket, AppWebSocket, True, ControllerWebSocket),
        (None, None, RouterWebSocket, AppWebSocket, True, RouterWebSocket),
        (None, None, None, AppWebSocket, True, AppWebSocket),
        (None, None, None, None, True, WebSocket),
        (None, None, None, None, False, WebSocket),
    ),
    ids=(
        "Custom class for all layers",
        "Custom class for all above handler layer",
        "Custom class for all above controller layer",
        "Custom class for all above router layer",
        "No custom class for layers",
        "No default class in app",
    ),
)
def test_websocket_class_resolution_of_layers(
    handler_websocket_class: Union[Type[WebSocket], None],
    controller_websocket_class: Union[Type[WebSocket], None],
    router_websocket_class: Union[Type[WebSocket], None],
    app_websocket_class: Union[Type[WebSocket], None],
    has_default_app_class: bool,
    expected: Type[WebSocket],
) -> None:
    class MyController(Controller):
        @websocket_listener("/")
        def handler(self, data: str) -> None:
            return

    if controller_websocket_class:
        MyController.websocket_class = ControllerWebSocket

    router = Router(path="/", route_handlers=[MyController])

    if router_websocket_class:
        router.websocket_class = router_websocket_class

    app = Litestar(route_handlers=[router])

    if app_websocket_class or not has_default_app_class:
        app.websocket_class = app_websocket_class  # type: ignore[assignment]

    route_handler = app.routes[0].route_handler  # type: ignore[union-attr]

    if handler_websocket_class:
        route_handler.websocket_class = handler_websocket_class  # type: ignore[union-attr]

    websocket_class = route_handler.resolve_websocket_class()  # type: ignore[union-attr]
    assert websocket_class is expected


@pytest.mark.parametrize(
    "handler_websocket_class, router_websocket_class, app_websocket_class, has_default_app_class, expected",
    (
        (HandlerWebSocket, RouterWebSocket, AppWebSocket, True, HandlerWebSocket),
        (None, RouterWebSocket, AppWebSocket, True, RouterWebSocket),
        (None, None, AppWebSocket, True, AppWebSocket),
        (None, None, None, True, WebSocket),
        (None, None, None, False, WebSocket),
    ),
    ids=(
        "Custom class for all layers",
        "Custom class for all above handler layer",
        "Custom class for all above router layer",
        "No custom class for layers",
        "No default class in app",
    ),
)
def test_listener_websocket_class_resolution_of_layers(
    handler_websocket_class: Union[Type[WebSocket], None],
    router_websocket_class: Union[Type[WebSocket], None],
    app_websocket_class: Union[Type[WebSocket], None],
    has_default_app_class: bool,
    expected: Type[WebSocket],
) -> None:
    class Handler(WebsocketListener):
        path = "/"
        websocket_class = handler_websocket_class

        def on_receive(self, data: str) -> str:  # pyright: ignore
            return data

    router = Router(path="/", route_handlers=[Handler])

    if router_websocket_class:
        router.websocket_class = router_websocket_class

    app = Litestar(route_handlers=[router])

    if app_websocket_class or not has_default_app_class:
        app.websocket_class = app_websocket_class  # type: ignore[assignment]

    route_handler = app.routes[0].route_handler  # type: ignore[union-attr]

    if handler_websocket_class:
        route_handler.websocket_class = handler_websocket_class  # type: ignore[union-attr]

    websocket_class = route_handler.resolve_websocket_class()  # type: ignore[union-attr]
    assert websocket_class is expected



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar.plugins import OpenAPISchemaPlugin
from litestar.typing import FieldDefinition

if TYPE_CHECKING:
    from litestar._openapi.schema_generation import SchemaCreator
    from litestar.openapi.spec import Schema


class TypedDictSchemaPlugin(OpenAPISchemaPlugin):
    def is_plugin_supported_field(self, field_definition: FieldDefinition) -> bool:
        return field_definition.is_typeddict_type

    def to_openapi_schema(self, field_definition: FieldDefinition, schema_creator: SchemaCreator) -> Schema:
        type_hints = field_definition.get_type_hints(include_extras=True, resolve_generics=True)

        return schema_creator.create_component_schema(
            field_definition,
            required=sorted(getattr(field_definition.type_, "__required_keys__", [])),
            property_fields={k: FieldDefinition.from_kwarg(v, k) for k, v in type_hints.items()},
        )



import sys
from dataclasses import dataclass
from datetime import date, datetime
from enum import Enum, auto
from typing import (  # type: ignore[attr-defined]
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypedDict,
    TypeVar,
    Union,
    _GenericAlias,  # pyright: ignore
)

import annotated_types
import msgspec
import pytest
from msgspec import Struct
from typing_extensions import Annotated, TypeAlias

from litestar import Controller, MediaType, get, post
from litestar._openapi.schema_generation.plugins import openapi_schema_plugins
from litestar._openapi.schema_generation.schema import (
    KWARG_DEFINITION_ATTRIBUTE_TO_OPENAPI_PROPERTY_MAP,
    SchemaCreator,
)
from litestar._openapi.schema_generation.utils import _get_normalized_schema_key, _type_or_first_not_none_inner_type
from litestar.app import DEFAULT_OPENAPI_CONFIG, Litestar
from litestar.di import Provide
from litestar.enums import ParamType
from litestar.openapi.spec import ExternalDocumentation, OpenAPIType, Reference
from litestar.openapi.spec.example import Example
from litestar.openapi.spec.parameter import Parameter as OpenAPIParameter
from litestar.openapi.spec.schema import Schema
from litestar.pagination import ClassicPagination, CursorPagination, OffsetPagination
from litestar.params import KwargDefinition, Parameter, ParameterKwarg
from litestar.testing import create_test_client
from litestar.types.builtin_types import NoneType
from litestar.typing import FieldDefinition
from litestar.utils.helpers import get_name
from tests.helpers import get_schema_for_field_definition
from tests.models import DataclassPerson, DataclassPet

if TYPE_CHECKING:
    from types import ModuleType
    from typing import Callable

T = TypeVar("T")


def test_process_schema_result() -> None:
    test_str = "abc"
    kwarg_definition = ParameterKwarg(
        examples=[Example(value=1)],
        external_docs=ExternalDocumentation(url="https://example.com/docs"),
        content_encoding="utf-8",
        default=test_str,
        title=test_str,
        description=test_str,
        const=True,
        gt=1,
        ge=1,
        lt=1,
        le=1,
        multiple_of=1,
        min_items=1,
        max_items=1,
        min_length=1,
        max_length=1,
        pattern="^[a-z]$",
    )
    schema = get_schema_for_field_definition(
        FieldDefinition.from_annotation(annotation=str, kwarg_definition=kwarg_definition)
    )

    assert schema.title
    assert schema.const == test_str
    assert kwarg_definition.examples
    for signature_key, schema_key in KWARG_DEFINITION_ATTRIBUTE_TO_OPENAPI_PROPERTY_MAP.items():
        if schema_key == "examples":
            assert schema.examples == [kwarg_definition.examples[0].value]
        else:
            assert getattr(schema, schema_key) == getattr(kwarg_definition, signature_key)


def test_get_normalized_schema_key() -> None:
    class LocalClass(msgspec.Struct):
        id: str

    # replace each of the long strings with underscores with a tuple of strings split at each underscore
    assert (
        "tests",
        "unit",
        "test_openapi",
        "test_schema",
        "test_get_normalized_schema_key.LocalClass",
    ) == _get_normalized_schema_key(LocalClass)

    assert ("tests", "models", "DataclassPerson") == _get_normalized_schema_key(DataclassPerson)

    builtin_dict = Dict[str, List[int]]
    assert ("typing", "Dict[str, typing.List[int]]") == _get_normalized_schema_key(builtin_dict)

    builtin_with_custom = Dict[str, DataclassPerson]
    assert ("typing", "Dict[str, tests.models.DataclassPerson]") == _get_normalized_schema_key(builtin_with_custom)

    class LocalGeneric(Generic[T]):
        pass

    assert (
        "tests",
        "unit",
        "test_openapi",
        "test_schema",
        "test_get_normalized_schema_key.LocalGeneric",
    ) == _get_normalized_schema_key(LocalGeneric)

    generic_int = LocalGeneric[int]
    generic_str = LocalGeneric[str]

    assert (
        "tests",
        "unit",
        "test_openapi",
        "test_schema",
        "test_get_normalized_schema_key.LocalGeneric[int]",
    ) == _get_normalized_schema_key(generic_int)

    assert (
        "tests",
        "unit",
        "test_openapi",
        "test_schema",
        "test_get_normalized_schema_key.LocalGeneric[str]",
    ) == _get_normalized_schema_key(generic_str)

    assert _get_normalized_schema_key(generic_int) != _get_normalized_schema_key(generic_str)


def test_dependency_schema_generation() -> None:
    async def top_dependency(query_param: int) -> int:
        return query_param

    async def mid_level_dependency(header_param: str = Parameter(header="header_param", required=False)) -> int:
        return 5

    async def local_dependency(path_param: int, mid_level: int, top_level: int) -> int:
        return path_param + mid_level + top_level

    class MyController(Controller):
        path = "/test"
        dependencies = {"mid_level": Provide(mid_level_dependency)}

        @get(
            path="/{path_param:int}",
            dependencies={
                "summed": Provide(local_dependency),
            },
            media_type=MediaType.TEXT,
        )
        def test_function(self, summed: int, handler_param: int) -> str:
            return str(summed)

    with create_test_client(
        MyController,
        dependencies={"top_level": Provide(top_dependency)},
        openapi_config=DEFAULT_OPENAPI_CONFIG,
    ) as client:
        handler = client.app.openapi_schema.paths["/test/{path_param}"]
        data = {param.name: {"in": param.param_in, "required": param.required} for param in handler.get.parameters}
        assert data == {
            "path_param": {"in": ParamType.PATH, "required": True},
            "header_param": {"in": ParamType.HEADER, "required": False},
            "query_param": {"in": ParamType.QUERY, "required": True},
            "handler_param": {"in": ParamType.QUERY, "required": True},
        }


def test_get_schema_for_annotation_enum() -> None:
    class Opts(str, Enum):
        opt1 = "opt1"
        opt2 = "opt2"

    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(Opts))
    assert schema.enum == ["opt1", "opt2"]


ValueType: TypeAlias = Literal["a", "b", "c"]
ConstType: TypeAlias = Literal[1]


def test_handling_of_literals() -> None:
    @dataclass
    class DataclassWithLiteral:
        value: ValueType
        const: ConstType
        composite: Literal[ValueType, ConstType]

    schema = get_schema_for_field_definition(FieldDefinition.from_kwarg(name="", annotation=DataclassWithLiteral))

    assert isinstance(schema, Schema)
    assert schema.properties

    value = schema.properties["value"]
    assert isinstance(value, Schema)
    assert value.enum == ["a", "b", "c"]

    const = schema.properties["const"]
    assert isinstance(const, Schema)
    assert const.const == 1

    composite = schema.properties["composite"]
    assert isinstance(composite, Schema)
    assert composite.enum == ["a", "b", "c", 1]


def test_schema_hashing() -> None:
    schema = Schema(
        one_of=[
            Schema(type=OpenAPIType.STRING),
            Schema(type=OpenAPIType.NUMBER),
            Schema(type=OpenAPIType.OBJECT, properties={"key": Schema(type=OpenAPIType.STRING)}),
        ],
        examples=[None, [1, 2, 3]],
    )
    assert hash(schema)


def test_title_validation() -> None:
    # TODO: what is this actually testing?
    creator = SchemaCreator(plugins=openapi_schema_plugins)
    person_ref = creator.for_field_definition(FieldDefinition.from_kwarg(name="Person", annotation=DataclassPerson))
    pet_ref = creator.for_field_definition(FieldDefinition.from_kwarg(name="Pet", annotation=DataclassPet))
    assert isinstance(person_ref, Reference)
    assert isinstance(pet_ref, Reference)
    assert isinstance(creator.schema_registry.from_reference(person_ref).schema, Schema)
    assert isinstance(creator.schema_registry.from_reference(pet_ref).schema, Schema)


@pytest.mark.parametrize("with_future_annotations", [True, False])
def test_create_schema_for_dataclass_with_annotated_model_attribute(
    with_future_annotations: bool, create_module: "Callable[[str], ModuleType]"
) -> None:
    """Test that a model with an annotated attribute is correctly handled."""
    module = create_module(
        f"""
{'from __future__ import annotations' if with_future_annotations else ''}
from typing_extensions import Annotated
from dataclasses import dataclass

@dataclass
class Foo:
    foo: Annotated[int, "Foo description"]
"""
    )
    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(module.Foo))
    assert schema.properties and "foo" in schema.properties


@pytest.mark.parametrize("with_future_annotations", [True, False])
def test_create_schema_for_typedict_with_annotated_required_and_not_required_model_attributes(
    with_future_annotations: bool, create_module: "Callable[[str], ModuleType]"
) -> None:
    """Test that a model with an annotated attribute is correctly handled."""
    module = create_module(
        f"""
{'from __future__ import annotations' if with_future_annotations else ''}
from typing_extensions import Annotated, Required, NotRequired
from typing import TypedDict

class Foo(TypedDict):
    foo: Annotated[int, "Foo description"]
    bar: Annotated[Required[int], "Bar description"]
    baz: Annotated[NotRequired[int], "Baz description"]
"""
    )
    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(module.Foo))
    assert schema.properties and all(key in schema.properties for key in ("foo", "bar", "baz"))


def test_create_schema_from_msgspec_annotated_type() -> None:
    class Lookup(msgspec.Struct):
        id: Annotated[str, msgspec.Meta(max_length=16, examples=["example"], description="description", title="title")]

    schema = get_schema_for_field_definition(FieldDefinition.from_kwarg(name="Lookup", annotation=Lookup))

    assert schema.properties["id"].type == OpenAPIType.STRING  # type: ignore[index, union-attr]
    assert schema.properties["id"].examples == ["example"]  # type: ignore[index, union-attr]
    assert schema.properties["id"].description == "description"  # type: ignore[index]
    assert schema.properties["id"].title == "title"  # type: ignore[index, union-attr]
    assert schema.properties["id"].max_length == 16  # type: ignore[index, union-attr]
    assert schema.required == ["id"]


def test_annotated_types() -> None:
    historical_date = date(year=1980, day=1, month=1)
    today = date.today()

    @dataclass
    class MyDataclass:
        constrained_int: Annotated[int, annotated_types.Gt(1), annotated_types.Lt(10)]
        constrained_float: Annotated[float, annotated_types.Ge(1), annotated_types.Le(10)]
        constrained_date: Annotated[date, annotated_types.Interval(gt=historical_date, lt=today)]
        constrained_lower_case: Annotated[str, annotated_types.LowerCase]
        constrained_upper_case: Annotated[str, annotated_types.UpperCase]
        constrained_is_ascii: Annotated[str, annotated_types.IsAscii]
        constrained_is_digit: Annotated[str, annotated_types.IsDigits]

    schema = get_schema_for_field_definition(FieldDefinition.from_kwarg(name="MyDataclass", annotation=MyDataclass))

    assert schema.properties["constrained_int"].exclusive_minimum == 1  # type: ignore[index, union-attr]
    assert schema.properties["constrained_int"].exclusive_maximum == 10  # type: ignore[index, union-attr]
    assert schema.properties["constrained_float"].minimum == 1  # type: ignore[index, union-attr]
    assert schema.properties["constrained_float"].maximum == 10  # type: ignore[index, union-attr]
    assert datetime.utcfromtimestamp(schema.properties["constrained_date"].exclusive_minimum) == datetime.fromordinal(  # type: ignore[arg-type, index, union-attr]
        historical_date.toordinal()
    )
    assert datetime.utcfromtimestamp(schema.properties["constrained_date"].exclusive_maximum) == datetime.fromordinal(  # type: ignore[arg-type, index, union-attr]
        today.toordinal()
    )
    assert schema.properties["constrained_lower_case"].description == "must be in lower case"  # type: ignore[index]
    assert schema.properties["constrained_upper_case"].description == "must be in upper case"  # type: ignore[index]
    assert schema.properties["constrained_is_ascii"].pattern == "[[:ascii:]]"  # type: ignore[index, union-attr]
    assert schema.properties["constrained_is_digit"].pattern == "[[:digit:]]"  # type: ignore[index, union-attr]


def test_literal_enums() -> None:
    class Foo(Enum):
        A = auto()
        B = auto()

    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(List[Literal[Foo.A]]))
    assert isinstance(schema.items, Schema)
    assert schema.items.const == 1


@dataclass
class DataclassGeneric(Generic[T]):
    foo: T
    optional_foo: Optional[T]
    annotated_foo: Annotated[T, object()]


class MsgspecGeneric(Struct, Generic[T]):
    foo: T
    optional_foo: Optional[T]
    annotated_foo: Annotated[T, object()]


annotations: List[type] = [DataclassGeneric[int], MsgspecGeneric[int]]

# Generic TypedDict was only supported from 3.11 onwards
if sys.version_info >= (3, 11):

    class TypedDictGeneric(TypedDict, Generic[T]):
        foo: T
        optional_foo: Optional[T]
        annotated_foo: Annotated[T, object()]

    annotations.append(TypedDictGeneric[int])


@pytest.mark.parametrize("cls", annotations)
def test_schema_generation_with_generic_classes(cls: Any) -> None:
    expected_foo_schema = Schema(type=OpenAPIType.INTEGER)
    expected_optional_foo_schema = Schema(one_of=[Schema(type=OpenAPIType.NULL), Schema(type=OpenAPIType.INTEGER)])

    properties = get_schema_for_field_definition(
        FieldDefinition.from_kwarg(name=get_name(cls), annotation=cls)
    ).properties
    assert properties
    assert properties["foo"] == expected_foo_schema
    assert properties["annotated_foo"] == expected_foo_schema
    assert properties["optional_foo"] == expected_optional_foo_schema


B = TypeVar("B", bound=int)
C = TypeVar("C", int, str)


@dataclass
class ConstrainedGenericDataclass(Generic[T, B, C]):
    bound: B
    constrained: C
    union: Union[T, bool]
    union_constrained: Union[C, bool]
    union_bound: Union[B, bool]


def test_schema_generation_with_generic_classes_constrained() -> None:
    cls = ConstrainedGenericDataclass
    properties = get_schema_for_field_definition(
        FieldDefinition.from_kwarg(name=cls.__name__, annotation=cls)
    ).properties

    assert properties
    assert properties["bound"] == Schema(type=OpenAPIType.INTEGER)
    assert properties["constrained"] == Schema(
        one_of=[Schema(type=OpenAPIType.INTEGER), Schema(type=OpenAPIType.STRING)]
    )
    assert properties["union"] == Schema(one_of=[Schema(type=OpenAPIType.OBJECT), Schema(type=OpenAPIType.BOOLEAN)])
    assert properties["union_constrained"] == Schema(
        one_of=[Schema(type=OpenAPIType.INTEGER), Schema(type=OpenAPIType.STRING), Schema(type=OpenAPIType.BOOLEAN)]
    )
    assert properties["union_bound"] == Schema(
        one_of=[Schema(type=OpenAPIType.INTEGER), Schema(type=OpenAPIType.BOOLEAN)]
    )


@pytest.mark.parametrize(
    "annotation",
    (
        ClassicPagination[DataclassGeneric[int]],
        OffsetPagination[DataclassGeneric[int]],
        CursorPagination[int, DataclassGeneric[int]],
    ),
)
def test_schema_generation_with_pagination(annotation: Any) -> None:
    expected_foo_schema = Schema(type=OpenAPIType.INTEGER)
    expected_optional_foo_schema = Schema(one_of=[Schema(type=OpenAPIType.NULL), Schema(type=OpenAPIType.INTEGER)])

    properties = get_schema_for_field_definition(FieldDefinition.from_annotation(annotation).inner_types[-1]).properties

    assert properties
    assert properties["foo"] == expected_foo_schema
    assert properties["annotated_foo"] == expected_foo_schema
    assert properties["optional_foo"] == expected_optional_foo_schema


def test_schema_generation_with_ellipsis() -> None:
    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(Tuple[int, ...]))
    assert isinstance(schema.items, Schema)
    assert schema.items.type == OpenAPIType.INTEGER


def test_schema_tuple_with_union() -> None:
    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(Tuple[int, Union[int, str]]))
    assert isinstance(schema.items, Schema)
    assert schema.items.one_of == [
        Schema(type=OpenAPIType.INTEGER),
        Schema(one_of=[Schema(type=OpenAPIType.INTEGER), Schema(type=OpenAPIType.STRING)]),
    ]


def test_optional_enum() -> None:
    class Foo(Enum):
        A = 1
        B = 2

    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(Optional[Foo]))
    assert schema.type is not None
    assert set(schema.type) == {OpenAPIType.INTEGER, OpenAPIType.NULL}
    assert schema.enum == [1, 2, None]


def test_optional_literal() -> None:
    schema = get_schema_for_field_definition(FieldDefinition.from_annotation(Optional[Literal[1]]))
    assert schema.type is not None
    assert set(schema.type) == {OpenAPIType.INTEGER, OpenAPIType.NULL}
    assert schema.enum == [1, None]


@pytest.mark.parametrize(
    ("in_type", "out_type"),
    [
        (FieldDefinition.from_annotation(Optional[int]), int),
        (FieldDefinition.from_annotation(Union[None, int]), int),
        (FieldDefinition.from_annotation(int), int),
        # hack to create a union of NoneType, NoneType to hit a branch for coverage
        (FieldDefinition.from_annotation(_GenericAlias(Union, (NoneType, NoneType))), ValueError),
    ],
)
def test_type_or_first_not_none_inner_type_utility(in_type: Any, out_type: Any) -> None:
    if out_type is ValueError:
        with pytest.raises(out_type):
            _type_or_first_not_none_inner_type(in_type)
    else:
        assert _type_or_first_not_none_inner_type(in_type) == out_type


def test_not_generating_examples_property() -> None:
    with_examples = SchemaCreator(generate_examples=True)
    without_examples = with_examples.not_generating_examples
    assert without_examples.generate_examples is False


def test_process_schema_result_with_unregistered_object_schema() -> None:
    """This test ensures that if a schema is created for an object and not registered in the schema registry, the
    schema is returned as-is, and not referenced.
    """
    schema = Schema(title="has title", type=OpenAPIType.OBJECT)
    field_definition = FieldDefinition.from_annotation(dict)
    assert SchemaCreator().process_schema_result(field_definition, schema) is schema


@pytest.mark.parametrize("base_type", [msgspec.Struct, TypedDict, dataclass])
def test_type_union(base_type: type) -> None:
    if base_type is dataclass:  # type: ignore[comparison-overlap]

        @dataclass
        class ModelA:  # pyright: ignore
            pass

        @dataclass
        class ModelB:  # pyright: ignore
            pass

    else:

        class ModelA(base_type):  # type: ignore[no-redef, misc]
            pass

        class ModelB(base_type):  # type: ignore[no-redef, misc]
            pass

    schema = get_schema_for_field_definition(
        FieldDefinition.from_kwarg(name="Lookup", annotation=Union[ModelA, ModelB])
    )
    assert schema.one_of == [
        Reference(ref="#/components/schemas/tests_unit_test_openapi_test_schema_test_type_union.ModelA"),
        Reference(ref="#/components/schemas/tests_unit_test_openapi_test_schema_test_type_union.ModelB"),
    ]


@pytest.mark.parametrize("base_type", [msgspec.Struct, TypedDict, dataclass])
def test_type_union_with_none(base_type: type) -> None:
    # https://github.com/litestar-org/litestar/issues/2971
    if base_type is dataclass:  # type: ignore[comparison-overlap]

        @dataclass
        class ModelA:  # pyright: ignore
            pass

        @dataclass
        class ModelB:  # pyright: ignore
            pass

    else:

        class ModelA(base_type):  # type: ignore[no-redef, misc]
            pass

        class ModelB(base_type):  # type: ignore[no-redef, misc]
            pass

    schema = get_schema_for_field_definition(
        FieldDefinition.from_kwarg(name="Lookup", annotation=Union[ModelA, ModelB, None])
    )
    assert schema.one_of == [
        Schema(type=OpenAPIType.NULL),
        Reference(ref="#/components/schemas/tests_unit_test_openapi_test_schema_test_type_union_with_none.ModelA"),
        Reference("#/components/schemas/tests_unit_test_openapi_test_schema_test_type_union_with_none.ModelB"),
    ]


def test_default_only_on_field_definition() -> None:
    field_definition = FieldDefinition.from_annotation(int, default=10)
    assert field_definition.kwarg_definition is None

    schema = get_schema_for_field_definition(field_definition)
    assert schema.default == 10


def test_default_not_provided_for_kwarg_but_for_field() -> None:
    field_definition = FieldDefinition.from_annotation(int, default=10, kwarg_definition=KwargDefinition())
    schema = get_schema_for_field_definition(field_definition)

    assert schema.default == 10


def test_routes_with_different_path_param_types_get_merged() -> None:
    # https://github.com/litestar-org/litestar/issues/2700
    @get("/{param:int}")
    async def get_handler(param: int) -> None:
        pass

    @post("/{param:str}")
    async def post_handler(param: str) -> None:
        pass

    app = Litestar([get_handler, post_handler])
    assert app.openapi_schema.paths
    paths = app.openapi_schema.paths["/{param}"]
    assert paths.get is not None
    assert paths.post is not None


def test_unconsumed_path_parameters_are_documented() -> None:
    # https://github.com/litestar-org/litestar/issues/3290
    @get("/{param:str}")
    async def handler() -> None:
        pass

    app = Litestar([handler])
    params = app.openapi_schema.paths["/{param}"].get.parameters  # type: ignore[index, union-attr]
    assert params
    assert len(params) == 1
    param = params[0]
    assert isinstance(param, OpenAPIParameter)
    assert param.name == "param"
    assert param.required is True
    assert param.param_in is ParamType.PATH



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar._openapi.schema_generation import SchemaCreator
from litestar.enums import RequestEncodingType
from litestar.openapi.spec.media_type import OpenAPIMediaType
from litestar.openapi.spec.request_body import RequestBody
from litestar.params import BodyKwarg

__all__ = ("create_request_body",)


if TYPE_CHECKING:
    from litestar._openapi.datastructures import OpenAPIContext
    from litestar.dto import AbstractDTO
    from litestar.typing import FieldDefinition


def create_request_body(
    context: OpenAPIContext,
    handler_id: str,
    resolved_data_dto: type[AbstractDTO] | None,
    data_field: FieldDefinition,
) -> RequestBody:
    """Create a RequestBody instance for the given route handler's data field.

    Args:
        context: The OpenAPIContext instance.
        handler_id: The handler id.
        resolved_data_dto: The resolved data dto.
        data_field: The data field.

    Returns:
        A RequestBody instance.
    """
    media_type: RequestEncodingType | str = RequestEncodingType.JSON
    schema_creator = SchemaCreator.from_openapi_context(context, prefer_alias=True)
    if isinstance(data_field.kwarg_definition, BodyKwarg) and data_field.kwarg_definition.media_type:
        media_type = data_field.kwarg_definition.media_type

    if resolved_data_dto:
        schema = resolved_data_dto.create_openapi_schema(
            field_definition=data_field,
            handler_id=handler_id,
            schema_creator=schema_creator,
        )
    else:
        schema = schema_creator.for_field_definition(data_field)

    return RequestBody(required=True, content={media_type: OpenAPIMediaType(schema=schema)})



from typing import TYPE_CHECKING, Any, Dict, Optional
from uuid import uuid4

import msgspec
from starlette.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_401_UNAUTHORIZED,
)

from litestar import Litestar, Request, delete, get, post
from litestar.middleware.session.server_side import (
    ServerSideSessionBackend,
    ServerSideSessionConfig,
)
from litestar.security.session_auth import SessionAuth
from litestar.testing import create_test_client
from tests.models import User, UserFactory

if TYPE_CHECKING:
    from litestar.connection import ASGIConnection

user_instance = UserFactory.build()


def retrieve_user_handler(session_data: Dict[str, Any], _: "ASGIConnection") -> Optional[User]:
    if session_data["id"] == str(user_instance.id):
        return User(**session_data)
    return None


def test_authentication(session_backend_config_memory: ServerSideSessionConfig) -> None:
    session_auth = SessionAuth[Any, ServerSideSessionBackend](
        retrieve_user_handler=retrieve_user_handler,
        exclude=["login"],
        session_backend_config=session_backend_config_memory,
    )

    @post("/login")
    def login_handler(request: "Request[Any, Any, Any]", data: User) -> None:
        request.set_session(msgspec.to_builtins(data))

    @delete("/user/{user_id:str}")
    def delete_user_handler(request: "Request[User, Any, Any]") -> None:
        request.clear_session()

    @get("/user/{user_id:str}")
    def get_user_handler(request: "Request[User, Any, Any]") -> User:
        return request.user

    with create_test_client(
        route_handlers=[login_handler, delete_user_handler, get_user_handler],
        on_app_init=[session_auth.on_app_init],
    ) as client:
        response = client.get(f"user/{user_instance.id}")
        assert response.status_code == HTTP_401_UNAUTHORIZED, response.json()

        response = client.post("/login", json={"id": str(user_instance.id), "name": user_instance.name})
        assert response.status_code == HTTP_201_CREATED, response.json()

        response = client.get(f"user/{user_instance.id}")
        assert response.status_code == HTTP_200_OK, response.json()

        response = client.delete(f"user/{user_instance.id}")
        assert response.status_code == HTTP_204_NO_CONTENT, response.json()

        response = client.get(f"user/{user_instance.id}")
        assert response.status_code == HTTP_401_UNAUTHORIZED, response.json()

        response = client.post("/login", json={"id": str(uuid4()), "name": user_instance.name})
        assert response.status_code == HTTP_201_CREATED, response.json()

        response = client.get(f"user/{user_instance.id}")
        assert response.status_code == HTTP_401_UNAUTHORIZED, response.json()


def test_session_auth_openapi(session_backend_config_memory: "ServerSideSessionConfig") -> None:
    session_auth = SessionAuth[Any, ServerSideSessionBackend](
        retrieve_user_handler=retrieve_user_handler,
        session_backend_config=session_backend_config_memory,
    )
    app = Litestar(on_app_init=[session_auth.on_app_init])
    assert app.openapi_schema.to_schema() == {
        "openapi": "3.1.0",
        "info": {"title": "Litestar API", "version": "1.0.0"},
        "servers": [{"url": "/"}],
        "paths": {},
        "components": {
            "schemas": {},
            "securitySchemes": {
                "sessionCookie": {
                    "type": "apiKey",
                    "description": "Session cookie authentication.",
                    "name": session_backend_config_memory.key,
                    "in": "cookie",
                }
            },
        },
        "security": [{"sessionCookie": []}],
    }



import secrets
from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

from litestar.testing import TestClient


@pytest.fixture(autouse=True)
def _chdir(tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
    monkeypatch.chdir(tmp_path)


@pytest.fixture()
def assets_file(tmp_path: Path) -> str:
    content = secrets.token_hex()
    assets_path = tmp_path / "assets"
    assets_path.mkdir()
    assets_path.joinpath("test.txt").write_text(content)
    return content


def test_custom_router() -> None:
    from docs.examples.static_files import custom_router  # noqa: F401


def test_full_example() -> None:
    from docs.examples.static_files import full_example

    with TestClient(full_example.app) as client:
        assert client.get("/static/hello.txt").text == "Hello, world!"


def test_html_mode() -> None:
    from docs.examples.static_files import html_mode

    with TestClient(html_mode.app) as client:
        assert client.get("/").text == "<strong>Hello, world!</strong>"
        assert client.get("/index.html").text == "<strong>Hello, world!</strong>"
        assert client.get("/something").text == "<h1>Not found</h1>"


def test_passing_options() -> None:
    from docs.examples.static_files import passing_options  # noqa: F401


def test_route_reverse(capsys) -> None:
    from docs.examples.static_files import route_reverse  # noqa: F401

    assert capsys.readouterr().out.strip() == "/static/some_file.txt"


def test_send_as_attachment(tmp_path: Path, assets_file: str) -> None:
    from docs.examples.static_files import send_as_attachment

    with TestClient(send_as_attachment.app) as client:
        res = client.get("/static/test.txt")
        assert res.text == assets_file
        assert res.headers["content-disposition"].startswith("attachment")


def test_upgrade_from_static(tmp_path: Path, assets_file: str) -> None:
    from docs.examples.static_files import upgrade_from_static_1, upgrade_from_static_2

    for app in [upgrade_from_static_1.app, upgrade_from_static_2.app]:
        with TestClient(app) as client:
            res = client.get("/static/test.txt")
            assert res.text == assets_file



"""Application ORM configuration."""

from __future__ import annotations

try:
    # v0.6.0+
    from advanced_alchemy._listeners import touch_updated_timestamp  # pyright: ignore
except ImportError:
    from advanced_alchemy.base import touch_updated_timestamp  # type: ignore[no-redef,attr-defined]

from advanced_alchemy.base import (
    AuditColumns,
    BigIntAuditBase,
    BigIntBase,
    BigIntPrimaryKey,
    CommonTableAttributes,
    ModelProtocol,
    UUIDAuditBase,
    UUIDBase,
    UUIDPrimaryKey,
    create_registry,
    orm_registry,
)

__all__ = (
    "AuditColumns",
    "BigIntAuditBase",
    "BigIntBase",
    "BigIntPrimaryKey",
    "CommonTableAttributes",
    "create_registry",
    "ModelProtocol",
    "touch_updated_timestamp",
    "UUIDAuditBase",
    "UUIDBase",
    "UUIDPrimaryKey",
    "orm_registry",
)



from __future__ import annotations

from litestar.handlers.websocket_handlers.listener import (
    WebsocketListener,
    WebsocketListenerRouteHandler,
    websocket_listener,
)
from litestar.handlers.websocket_handlers.route_handler import WebsocketRouteHandler, websocket

__all__ = (
    "WebsocketListener",
    "WebsocketListenerRouteHandler",
    "WebsocketRouteHandler",
    "websocket",
    "websocket_listener",
)



from __future__ import annotations

import atexit
from logging import Handler, LogRecord, StreamHandler
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import Any

from litestar.logging._utils import resolve_handlers

__all__ = ("LoggingQueueListener", "QueueListenerHandler")


class LoggingQueueListener(QueueListener):
    """Custom ``QueueListener`` which starts and stops the listening process."""

    def __init__(self, queue: Queue[LogRecord], *handlers: Handler, respect_handler_level: bool = False) -> None:
        """Initialize ``LoggingQueueListener``.

        Args:
            queue: The queue to send messages to
            *handlers: A list of handlers which will handle entries placed on the queue
            respect_handler_level: If ``respect_handler_level`` is ``True``, a handler's level is respected (compared with the level for the message) when deciding whether to pass messages to that handler
        """
        super().__init__(queue, *handlers, respect_handler_level=respect_handler_level)
        self.start()
        atexit.register(self.stop)


class QueueListenerHandler(QueueHandler):
    """Configure queue listener and handler to support non-blocking logging configuration.

    .. caution::

        This handler doesn't work with Python >= 3.12 and ``logging.config.dictConfig``. It might
        be deprecated in the future. Please use ``logging.QueueHandler`` instead.
    """

    def __init__(self, handlers: list[Any] | None = None) -> None:
        """Initialize ``QueueListenerHandler``.

        Args:
            handlers: Optional 'ConvertingList'
        """
        super().__init__(Queue(-1))
        handlers = resolve_handlers(handlers) if handlers else [StreamHandler()]
        self.listener = LoggingQueueListener(self.queue, *handlers)  # type: ignore[arg-type]



import pytest

from litestar import Litestar, get
from litestar.exceptions import ImproperlyConfiguredException


async def first_method(query_param: int) -> int:
    return query_param


async def second_method(path_param: str) -> str:
    return path_param


def test_dependency_validation() -> None:
    @get(
        path="/{path_param:int}",
        dependencies={"first": first_method, "second": second_method},
    )
    def handler(first: int, second: str, third: int) -> None:
        pass

    with pytest.raises(ImproperlyConfiguredException):
        Litestar(route_handlers=[handler], dependencies={"third": first_method})



from __future__ import annotations

from enum import Enum
from pathlib import Path

import pytest

from litestar import Request, get
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.contrib.mako import MakoTemplateEngine
from litestar.contrib.minijinja import MiniJinjaTemplateEngine
from litestar.plugins.flash import FlashConfig, FlashPlugin, flash
from litestar.response import Template
from litestar.template import TemplateConfig, TemplateEngineProtocol
from litestar.testing import create_test_client

text_html_jinja = """{% for message in get_flashes() %}<span class="{{ message.category }}">{{ message.message }}</span>{% endfor %}"""
text_html_mako = """<% messages = get_flashes() %>\\
% for m in messages:
<span class="${m['category']}">${m['message']}</span>\\
% endfor
"""


class CustomCategory(str, Enum):
    custom1 = "1"
    custom2 = "2"
    custom3 = "3"


class FlashCategory(str, Enum):
    info = "INFO"
    error = "ERROR"
    warning = "WARNING"
    success = "SUCCESS"


@pytest.mark.parametrize(
    "engine, template_str",
    (
        (JinjaTemplateEngine, text_html_jinja),
        (MakoTemplateEngine, text_html_mako),
        (MiniJinjaTemplateEngine, text_html_jinja),
    ),
    ids=("jinja", "mako", "minijinja"),
)
@pytest.mark.parametrize(
    "category_enum",
    (CustomCategory, FlashCategory),
    ids=("custom_category", "flash_category"),
)
def test_flash_plugin(
    tmp_path: Path,
    engine: type[TemplateEngineProtocol],
    template_str: str,
    category_enum: Enum,
) -> None:
    Path(tmp_path / "flash.html").write_text(template_str)
    text_expected = "".join(
        [f'<span class="{category.value}">message {category.value}</span>' for category in category_enum]  # type: ignore[attr-defined]
    )

    @get("/flash")
    def flash_handler(request: Request) -> Template:
        for category in category_enum:  # type: ignore[attr-defined]
            flash(request, f"message {category.value}", category=category.value)
        return Template("flash.html")

    template_config: TemplateConfig = TemplateConfig(
        directory=Path(tmp_path),
        engine=engine,
    )
    with create_test_client(
        [flash_handler],
        template_config=template_config,
        plugins=[FlashPlugin(config=FlashConfig(template_config=template_config))],
    ) as client:
        r = client.get("/flash")
        assert r.status_code == 200
        assert r.text == text_expected



from uuid import uuid4

from litestar import Controller, HttpMethod, Litestar, Response, Router, get
from litestar.datastructures import Cookie
from litestar.status_codes import HTTP_200_OK
from litestar.testing import create_test_client


def test_response_cookies() -> None:
    router_first = Cookie(key="second", value="1")
    router_second = Cookie(key="third", value="2")
    controller_first = Cookie(key="first", value="3")
    controller_second = Cookie(key="second", value="4")
    app_first = Cookie(key="first", value="5")
    app_second = Cookie(key="fourth", value="6")
    local_first = Cookie(key="first", value="7")

    test_path = "/test"

    class MyController(Controller):
        path = test_path
        response_cookies = [controller_first, controller_second]

        @get(
            path="/{path_param:str}",
            response_cookies=[local_first],
        )
        def test_method(self) -> None:
            pass

    first_router = Router(path="/users", response_cookies=[router_first, router_second], route_handlers=[MyController])

    second_router = Router(path="/external", response_cookies=[Cookie(key="external", value="nope")], route_handlers=[])

    app = Litestar(
        openapi_config=None,
        response_cookies=[app_first, app_second],
        route_handlers=[first_router, second_router],
    )
    route_handler, _ = app.routes[0].route_handler_map[HttpMethod.GET]  # type: ignore[union-attr]
    response_cookies = {cookie.key: cookie.value for cookie in route_handler.resolve_response_cookies()}
    assert response_cookies["first"] == local_first.value
    assert response_cookies["second"] == controller_second.value
    assert response_cookies["third"] == router_second.value
    assert response_cookies["fourth"] == app_second.value
    assert "external" not in response_cookies


def test_response_cookies_mapping() -> None:
    @get(response_cookies={"foo": "bar"})
    def handler_one() -> None:
        pass

    @get(response_cookies=[Cookie(key="foo", value="bar")])
    def handler_two() -> None:
        pass

    assert handler_one.resolve_response_cookies() == handler_two.resolve_response_cookies()


def test_response_cookies_mapping_unresolved() -> None:
    # this should never happen, as there's no way to create this situation which type-checks.
    # we test for it nevertheless

    @get()
    def handler_one() -> None:
        pass

    handler_one.response_cookies = {"foo": "bar"}  # type: ignore[assignment]

    assert handler_one.resolve_response_cookies() == frozenset([Cookie(key="foo", value="bar")])


def test_response_cookie_rendering() -> None:
    @get(
        "/",
        response_cookies=[Cookie(key="test", value="123")],
    )
    def test_method() -> None:
        return None

    with create_test_client(test_method) as client:
        response = client.get("/")
        assert response.headers["Set-Cookie"] == "test=123; Path=/; SameSite=lax"


def test_response_cookie_documentation_only_not_rendering() -> None:
    @get(
        "/",
        response_cookies=[
            Cookie(
                key="my-cookie",
                description="my-cookie documentations",
                documentation_only=True,
            )
        ],
    )
    def test_method() -> None:
        return None

    with create_test_client(test_method) as client:
        response = client.get("/")
        assert "Set-Cookie" not in response.headers


def test_response_cookie_documentation_only_not_producing_second_header() -> None:
    # https://github.com/litestar-org/litestar/issues/870
    def after_request(response: Response) -> Response:
        response.set_cookie("my-cookie", "123")
        return response

    @get(
        "/",
        response_cookies=[
            Cookie(
                key="my-cookie",
                description="my-cookie documentations",
                documentation_only=True,
            )
        ],
    )
    def test_method() -> None:
        return None

    with create_test_client(test_method, after_request=after_request) as client:
        response = client.get("/")
        assert response.headers["Set-Cookie"] == "my-cookie=123; Path=/; SameSite=lax"
        assert len(response.headers.get_list("Set-Cookie")) == 1


def test_response_cookie_is_always_set() -> None:
    # https://github.com/litestar-org/litestar/issues/888
    @get(path="/set-cookie")
    def set_cookie_handler() -> Response[None]:
        return Response(
            content=None,
            cookies=[
                Cookie(
                    key="test",
                    value=str(uuid4()),
                    expires=10,
                )
            ],
        )

    with create_test_client([set_cookie_handler]) as client:
        response = client.get("/set-cookie")
        assert response.status_code == HTTP_200_OK
        assert response.cookies.get("test")
        cookie = response.cookies.get("test")
        client.cookies.clear()
        response = client.get("/set-cookie")
        assert response.status_code == HTTP_200_OK
        assert response.cookies.get("test")
        assert cookie != response.cookies.get("test")



"""DTO backends do the heavy lifting of decoding and validating raw bytes into domain models, and
back again, to bytes.
"""

from __future__ import annotations

import re
import textwrap
from contextlib import contextmanager, nullcontext
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    Generator,
    Mapping,
    Protocol,
    cast,
)

from msgspec import UNSET

from litestar.dto._backend import DTOBackend
from litestar.dto._types import (
    CollectionType,
    CompositeType,
    SimpleType,
    TransferDTOFieldDefinition,
    TransferType,
    UnionType,
)
from litestar.utils.helpers import unique_name_for_scope

if TYPE_CHECKING:
    from litestar.connection import ASGIConnection
    from litestar.dto import AbstractDTO
    from litestar.types.serialization import LitestarEncodableType
    from litestar.typing import FieldDefinition

__all__ = ("DTOCodegenBackend",)


class DTOCodegenBackend(DTOBackend):
    __slots__ = (
        "_transfer_to_dict",
        "_transfer_to_model_type",
        "_transfer_data_from_builtins",
        "_transfer_data_from_builtins_with_overrides",
        "_encode_data",
    )

    def __init__(
        self,
        dto_factory: type[AbstractDTO],
        field_definition: FieldDefinition,
        handler_id: str,
        is_data_field: bool,
        model_type: type[Any],
        wrapper_attribute_name: str | None,
    ) -> None:
        """Create dto backend instance.

        Args:
            dto_factory: The DTO factory class calling this backend.
            field_definition: Parsed type.
            handler_id: The name of the handler that this backend is for.
            is_data_field: Whether the field is a subclass of DTOData.
            model_type: Model type.
            wrapper_attribute_name: If the data that DTO should operate upon is wrapped in a generic datastructure,
              this is the name of the attribute that the data is stored in.
        """
        super().__init__(
            dto_factory=dto_factory,
            field_definition=field_definition,
            handler_id=handler_id,
            is_data_field=is_data_field,
            model_type=model_type,
            wrapper_attribute_name=wrapper_attribute_name,
        )
        self._transfer_to_dict = self._create_transfer_data_fn(
            destination_type=dict,
            field_definition=self.field_definition,
        )
        self._transfer_to_model_type = self._create_transfer_data_fn(
            destination_type=self.model_type,
            field_definition=self.field_definition,
        )
        self._transfer_data_from_builtins = self._create_transfer_data_fn(
            destination_type=self.model_type,
            field_definition=self.field_definition,
        )
        self._transfer_data_from_builtins_with_overrides = self._create_transfer_data_fn(
            destination_type=self.model_type,
            field_definition=self.field_definition,
        )
        self._encode_data = self._create_transfer_data_fn(
            destination_type=self.transfer_model_type,
            field_definition=self.field_definition,
        )

    def populate_data_from_builtins(self, builtins: Any, asgi_connection: ASGIConnection) -> Any:
        """Populate model instance from builtin types.

        Args:
            builtins: Builtin type.
            asgi_connection: The current ASGI Connection

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        if self.dto_data_type:
            return self.dto_data_type(
                backend=self,
                data_as_builtins=self._transfer_to_dict(self.parse_builtins(builtins, asgi_connection)),
            )
        return self.transfer_data_from_builtins(self.parse_builtins(builtins, asgi_connection))

    def transfer_data_from_builtins(self, builtins: Any) -> Any:
        """Populate model instance from builtin types.

        Args:
            builtins: Builtin type.

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        return self._transfer_data_from_builtins(builtins)

    def populate_data_from_raw(self, raw: bytes, asgi_connection: ASGIConnection) -> Any:
        """Parse raw bytes into instance of `model_type`.

        Args:
            raw: bytes
            asgi_connection: The current ASGI Connection

        Returns:
            Instance or collection of ``model_type`` instances.
        """
        if self.dto_data_type:
            return self.dto_data_type(
                backend=self,
                data_as_builtins=self._transfer_to_dict(self.parse_raw(raw, asgi_connection)),
            )
        return self._transfer_to_model_type(self.parse_raw(raw, asgi_connection))

    def encode_data(self, data: Any) -> LitestarEncodableType:
        """Encode data into a ``LitestarEncodableType``.

        Args:
            data: Data to encode.

        Returns:
            Encoded data.
        """
        if self.wrapper_attribute_name:
            wrapped_transfer = self._encode_data(getattr(data, self.wrapper_attribute_name))
            setattr(data, self.wrapper_attribute_name, wrapped_transfer)
            return cast("LitestarEncodableType", data)

        return cast("LitestarEncodableType", self._encode_data(data))

    def _create_transfer_data_fn(
        self,
        destination_type: type[Any],
        field_definition: FieldDefinition,
    ) -> Any:
        """Create instance or iterable of instances of ``destination_type``.

        Args:
            destination_type: the model type received by the DTO on type narrowing.
            field_definition: the parsed type that represents the handler annotation for which the DTO is being applied.

        Returns:
            Data parsed into ``destination_type``.
        """

        return TransferFunctionFactory.create_transfer_data(
            destination_type=destination_type,
            field_definitions=self.parsed_field_definitions,
            is_data_field=self.is_data_field,
            field_definition=field_definition,
        )


class FieldAccessManager(Protocol):
    def __call__(self, source_name: str, field_name: str, expect_optional: bool) -> ContextManager[str]: ...


class TransferFunctionFactory:
    def __init__(self, is_data_field: bool, nested_as_dict: bool) -> None:
        self.is_data_field = is_data_field
        self._fn_locals: dict[str, Any] = {
            "Mapping": Mapping,
            "UNSET": UNSET,
        }
        self._indentation = 1
        self._body = ""
        self.names: set[str] = set()
        self.nested_as_dict = nested_as_dict
        self._re_index_access = re.compile(r"\[['\"](\w+?)['\"]]")

    def _add_to_fn_globals(self, name: str, value: Any) -> str:
        unique_name = unique_name_for_scope(name, self._fn_locals)
        self._fn_locals[unique_name] = value
        return unique_name

    def _create_local_name(self, name: str) -> str:
        unique_name = unique_name_for_scope(name, self.names)
        self.names.add(unique_name)
        return unique_name

    def _make_function(
        self, source_value_name: str, return_value_name: str, fn_name: str = "func"
    ) -> Callable[[Any], Any]:
        """Wrap the current body contents in a function definition and turn it into a callable object"""
        source = f"def {fn_name}({source_value_name}):\n{self._body} return {return_value_name}"
        ctx: dict[str, Any] = {**self._fn_locals}
        exec(source, ctx)  # noqa: S102
        return ctx["func"]  # type: ignore[no-any-return]

    def _add_stmt(self, stmt: str) -> None:
        self._body += textwrap.indent(stmt + "\n", " " * self._indentation)

    @contextmanager
    def _start_block(self, expr: str | None = None) -> Generator[None, None, None]:
        """Start an indented block. If `expr` is given, use it as the "opening line"
        of the block.
        """
        if expr is not None:
            self._add_stmt(expr)
        self._indentation += 1
        yield
        self._indentation -= 1

    @contextmanager
    def _try_except_pass(self, exception: str) -> Generator[None, None, None]:
        """Enter a `try / except / pass` block. Content written while inside this context
        will go into the `try` block.
        """
        with self._start_block("try:"):
            yield
        with self._start_block(expr=f"except {exception}:"):
            self._add_stmt("pass")

    @contextmanager
    def _access_mapping_item(
        self, source_name: str, field_name: str, expect_optional: bool
    ) -> Generator[str, None, None]:
        """Enter a context within which an item of a mapping can be accessed safely,
        i.e. only if it is contained within that mapping.
        Yields an expression that accesses the mapping item. Content written while
        within this context can use this expression to access the desired value.
        """
        value_expr = f"{source_name}['{field_name}']"

        # if we expect an optional item, it's faster to check if it exists beforehand
        if expect_optional:
            with self._start_block(f"if '{field_name}' in {source_name}:"):
                yield value_expr
        # the happy path of a try/except will be faster than that, so we use that if
        # we expect a value
        else:
            with self._try_except_pass("KeyError"):
                yield value_expr

    @contextmanager
    def _access_attribute(self, source_name: str, field_name: str, expect_optional: bool) -> Generator[str, None, None]:
        """Enter a context within which an attribute of an object can be accessed
        safely, i.e. only if the object actually has the attribute.
        Yields an expression that retrieves the object attribute. Content written while
        within this context can use this expression to access the desired value.
        """

        value_expr = f"{source_name}.{field_name}"

        # if we expect an optional attribute it's faster to check with hasattr
        if expect_optional:
            with self._start_block(f"if hasattr({source_name}, '{field_name}'):"):
                yield value_expr
        # the happy path of a try/except will be faster than that, so we use that if
        # we expect a value
        else:
            with self._try_except_pass("AttributeError"):
                yield value_expr

    @classmethod
    def create_transfer_instance_data(
        cls,
        field_definitions: tuple[TransferDTOFieldDefinition, ...],
        destination_type: type[Any],
        is_data_field: bool,
    ) -> Callable[[Any], Any]:
        factory = cls(is_data_field=is_data_field, nested_as_dict=destination_type is dict)
        tmp_return_type_name = factory._create_local_name("tmp_return_type")
        source_instance_name = factory._create_local_name("source_instance")
        destination_type_name = factory._add_to_fn_globals("destination_type", destination_type)
        factory._create_transfer_instance_data(
            tmp_return_type_name=tmp_return_type_name,
            source_instance_name=source_instance_name,
            destination_type_name=destination_type_name,
            field_definitions=field_definitions,
            destination_type_is_dict=destination_type is dict,
        )
        return factory._make_function(source_value_name=source_instance_name, return_value_name=tmp_return_type_name)

    @classmethod
    def create_transfer_type_data(
        cls,
        transfer_type: TransferType,
        is_data_field: bool,
    ) -> Callable[[Any], Any]:
        factory = cls(is_data_field=is_data_field, nested_as_dict=False)
        tmp_return_type_name = factory._create_local_name("tmp_return_type")
        source_value_name = factory._create_local_name("source_value")
        factory._create_transfer_type_data_body(
            transfer_type=transfer_type,
            nested_as_dict=False,
            assignment_target=tmp_return_type_name,
            source_value_name=source_value_name,
        )
        return factory._make_function(source_value_name=source_value_name, return_value_name=tmp_return_type_name)

    @classmethod
    def create_transfer_data(
        cls,
        destination_type: type[Any],
        field_definitions: tuple[TransferDTOFieldDefinition, ...],
        is_data_field: bool,
        field_definition: FieldDefinition | None = None,
    ) -> Callable[[Any], Any]:
        if field_definition and field_definition.is_non_string_collection:
            factory = cls(
                is_data_field=is_data_field,
                nested_as_dict=False,
            )
            source_value_name = factory._create_local_name("source_value")
            return_value_name = factory._create_local_name("tmp_return_value")
            factory._create_transfer_data_body_nested(
                field_definitions=field_definitions,
                field_definition=field_definition,
                destination_type=destination_type,
                source_data_name=source_value_name,
                assignment_target=return_value_name,
            )
            return factory._make_function(source_value_name=source_value_name, return_value_name=return_value_name)

        return cls.create_transfer_instance_data(
            destination_type=destination_type,
            field_definitions=field_definitions,
            is_data_field=is_data_field,
        )

    def _create_transfer_data_body_nested(
        self,
        field_definition: FieldDefinition,
        field_definitions: tuple[TransferDTOFieldDefinition, ...],
        destination_type: type[Any],
        source_data_name: str,
        assignment_target: str,
    ) -> None:
        origin_name = self._add_to_fn_globals("origin", field_definition.instantiable_origin)
        transfer_func = TransferFunctionFactory.create_transfer_data(
            is_data_field=self.is_data_field,
            destination_type=destination_type,
            field_definition=field_definition.inner_types[0],
            field_definitions=field_definitions,
        )
        transfer_func_name = self._add_to_fn_globals("transfer_data", transfer_func)
        if field_definition.is_mapping:
            self._add_stmt(
                f"{assignment_target} = {origin_name}((key, {transfer_func_name}(item)) for key, item in {source_data_name}.items())"
            )
        else:
            self._add_stmt(
                f"{assignment_target} = {origin_name}({transfer_func_name}(item) for item in {source_data_name})"
            )

    def _create_transfer_instance_data(
        self,
        tmp_return_type_name: str,
        source_instance_name: str,
        destination_type_name: str,
        field_definitions: tuple[TransferDTOFieldDefinition, ...],
        destination_type_is_dict: bool,
    ) -> None:
        local_dict_name = self._create_local_name("unstructured_data")
        self._add_stmt(f"{local_dict_name} = {{}}")

        if field_definitions := tuple(f for f in field_definitions if self.is_data_field or not f.is_excluded):
            if len(field_definitions) > 1 and ("." in source_instance_name or "[" in source_instance_name):
                # If there's more than one field we have to access, we check if it is
                # nested. If it is nested, we assign it to a local variable to avoid
                # repeated lookups. This is only a small performance improvement for
                # regular attributes, but can be quite significant for properties or
                # other types of descriptors, where I/O may be involved, such as the
                # case for lazy loaded relationships in SQLAlchemy
                if "." in source_instance_name:
                    level_1, level_2 = source_instance_name.split(".", 1)
                else:
                    level_1, level_2, *_ = self._re_index_access.split(source_instance_name, maxsplit=1)

                new_source_instance_name = self._create_local_name(f"{level_1}_{level_2}")
                self._add_stmt(f"{new_source_instance_name} = {source_instance_name}")
                source_instance_name = new_source_instance_name

            for source_type in ("mapping", "object"):
                if source_type == "mapping":
                    block_expr = f"if isinstance({source_instance_name}, Mapping):"
                    access_item = self._access_mapping_item
                else:
                    block_expr = "else:"
                    access_item = self._access_attribute

                with self._start_block(expr=block_expr):
                    self._create_transfer_instance_data_inner(
                        local_dict_name=local_dict_name,
                        field_definitions=field_definitions,
                        access_field_safe=access_item,
                        source_instance_name=source_instance_name,
                    )

        # if the destination type is a dict we can reuse our temporary dictionary of
        # unstructured data as the "return value"
        if not destination_type_is_dict:
            self._add_stmt(f"{tmp_return_type_name} = {destination_type_name}(**{local_dict_name})")
        else:
            self._add_stmt(f"{tmp_return_type_name} = {local_dict_name}")

    def _create_transfer_instance_data_inner(
        self,
        *,
        local_dict_name: str,
        field_definitions: tuple[TransferDTOFieldDefinition, ...],
        access_field_safe: FieldAccessManager,
        source_instance_name: str,
    ) -> None:
        for field_definition in field_definitions:
            with access_field_safe(
                source_name=source_instance_name,
                field_name=field_definition.name,
                expect_optional=field_definition.is_partial or field_definition.is_optional,
            ) as source_value_expr:
                if self.is_data_field and field_definition.is_partial:
                    # we assign the source value to a name here, so we can skip
                    # getting it twice from the source instance
                    source_value_name = self._create_local_name("source_value")
                    self._add_stmt(f"{source_value_name} = {source_value_expr}")
                    ctx = self._start_block(f"if {source_value_name} is not UNSET:")
                else:
                    # in these cases, we only ever access the source value once, so
                    # we can skip assigning it
                    source_value_name = source_value_expr
                    ctx = nullcontext()  # type: ignore[assignment]
                with ctx:
                    self._create_transfer_type_data_body(
                        transfer_type=field_definition.transfer_type,
                        nested_as_dict=self.nested_as_dict,
                        source_value_name=source_value_name,
                        assignment_target=f"{local_dict_name}['{field_definition.name}']",
                    )

    def _create_transfer_type_data_body(
        self,
        transfer_type: TransferType,
        nested_as_dict: bool,
        source_value_name: str,
        assignment_target: str,
    ) -> None:
        if isinstance(transfer_type, SimpleType) and transfer_type.nested_field_info:
            if nested_as_dict:
                destination_type: Any = dict
            elif self.is_data_field:
                destination_type = transfer_type.field_definition.annotation
            else:
                destination_type = transfer_type.nested_field_info.model

            self._create_transfer_instance_data(
                field_definitions=transfer_type.nested_field_info.field_definitions,
                tmp_return_type_name=assignment_target,
                source_instance_name=source_value_name,
                destination_type_name=self._add_to_fn_globals("destination_type", destination_type),
                destination_type_is_dict=destination_type is dict,
            )
            return

        if isinstance(transfer_type, UnionType) and transfer_type.has_nested:
            self._create_transfer_nested_union_type_data(
                transfer_type=transfer_type,
                source_value_name=source_value_name,
                assignment_target=assignment_target,
            )
            return

        if isinstance(transfer_type, CollectionType):
            origin_name = self._add_to_fn_globals("origin", transfer_type.field_definition.instantiable_origin)
            if transfer_type.has_nested:
                transfer_type_data_fn = TransferFunctionFactory.create_transfer_type_data(
                    is_data_field=self.is_data_field, transfer_type=transfer_type.inner_type
                )
                transfer_type_data_name = self._add_to_fn_globals("transfer_type_data", transfer_type_data_fn)
                self._add_stmt(
                    f"{assignment_target} = {origin_name}({transfer_type_data_name}(item) for item in {source_value_name})"
                )
                return

            self._add_stmt(f"{assignment_target} = {origin_name}({source_value_name})")
            return

        self._add_stmt(f"{assignment_target} = {source_value_name}")

    def _create_transfer_nested_union_type_data(
        self,
        transfer_type: UnionType,
        source_value_name: str,
        assignment_target: str,
    ) -> None:
        for inner_type in transfer_type.inner_types:
            if isinstance(inner_type, CompositeType):
                continue

            if inner_type.nested_field_info:
                if self.is_data_field:
                    constraint_type = inner_type.nested_field_info.model
                    destination_type = inner_type.field_definition.annotation
                else:
                    constraint_type = inner_type.field_definition.annotation
                    destination_type = inner_type.nested_field_info.model

                constraint_type_name = self._add_to_fn_globals("constraint_type", constraint_type)
                destination_type_name = self._add_to_fn_globals("destination_type", destination_type)

                with self._start_block(f"if isinstance({source_value_name}, {constraint_type_name}):"):
                    self._create_transfer_instance_data(
                        destination_type_name=destination_type_name,
                        destination_type_is_dict=destination_type is dict,
                        field_definitions=inner_type.nested_field_info.field_definitions,
                        source_instance_name=source_value_name,
                        tmp_return_type_name=assignment_target,
                    )
                    return
        self._add_stmt(f"{assignment_target} = {source_value_name}")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, NamedTuple

from litestar.enums import ParamType
from litestar.params import ParameterKwarg

if TYPE_CHECKING:
    from litestar.typing import FieldDefinition

__all__ = ("ParameterDefinition", "create_parameter_definition", "merge_parameter_sets")


class ParameterDefinition(NamedTuple):
    """Tuple defining a kwarg representing a request parameter."""

    default: Any
    field_alias: str
    field_name: str
    is_required: bool
    is_sequence: bool
    param_type: ParamType


def create_parameter_definition(
    field_definition: FieldDefinition,
    field_name: str,
    path_parameters: set[str],
) -> ParameterDefinition:
    """Create a ParameterDefinition for the given FieldDefinition.

    Args:
        field_definition: FieldDefinition instance.
        field_name: The field's name.
        path_parameters: A set of path parameter names.

    Returns:
        A ParameterDefinition tuple.
    """
    default = field_definition.default if field_definition.has_default else None
    kwarg_definition = (
        field_definition.kwarg_definition if isinstance(field_definition.kwarg_definition, ParameterKwarg) else None
    )

    field_alias = kwarg_definition.query if kwarg_definition and kwarg_definition.query else field_name
    param_type = ParamType.QUERY

    if field_name in path_parameters:
        field_alias = field_name
        param_type = ParamType.PATH
    elif kwarg_definition and kwarg_definition.header:
        field_alias = kwarg_definition.header
        param_type = ParamType.HEADER
    elif kwarg_definition and kwarg_definition.cookie:
        field_alias = kwarg_definition.cookie
        param_type = ParamType.COOKIE

    return ParameterDefinition(
        param_type=param_type,
        field_name=field_name,
        field_alias=field_alias,
        default=default,
        is_required=field_definition.is_required
        and default is None
        and not field_definition.is_optional
        and not field_definition.is_any,
        is_sequence=field_definition.is_non_string_sequence,
    )


def merge_parameter_sets(first: set[ParameterDefinition], second: set[ParameterDefinition]) -> set[ParameterDefinition]:
    """Given two sets of parameter definitions, coming from different dependencies for example, merge them into a single
    set.
    """
    result: set[ParameterDefinition] = first.intersection(second)
    difference = first.symmetric_difference(second)
    for param in difference:
        # add the param if it's either required or no-other param in difference is the same but required
        if param.is_required or not any(p.field_alias == param.field_alias and p.is_required for p in difference):
            result.add(param)
    return result



from typing import Optional

import pytest

from litestar import Controller, Request, Response, Router, get
from litestar.testing import create_test_client
from litestar.types import AnyCallable, BeforeRequestHookHandler


def sync_before_request_handler_with_return_value(request: Request) -> dict:
    assert isinstance(request, Request)
    return {"hello": "moon"}


async def async_before_request_handler_with_return_value(request: Request) -> dict:
    assert isinstance(request, Request)
    return {"hello": "moon"}


def sync_before_request_handler_without_return_value(request: Request) -> None:
    assert isinstance(request, Request)


async def async_before_request_handler_without_return_value(request: Request) -> None:
    assert isinstance(request, Request)


def sync_after_request_handler(response: Response) -> Response:
    assert isinstance(response, Response)
    response.content = {"hello": "moon"}
    return response


async def async_after_request_handler(response: Response) -> Response:
    assert isinstance(response, Response)
    response.content = {"hello": "moon"}
    return response


@pytest.mark.parametrize(
    "before_request, expected",
    (
        (None, {"hello": "world"}),
        (sync_before_request_handler_with_return_value, {"hello": "moon"}),
        (async_before_request_handler_with_return_value, {"hello": "moon"}),
        (sync_before_request_handler_without_return_value, {"hello": "world"}),
        (async_before_request_handler_without_return_value, {"hello": "world"}),
    ),
)
def test_before_request_handler_called(before_request: Optional[AnyCallable], expected: dict) -> None:
    @get(before_request=before_request)
    def handler() -> dict:
        return {"hello": "world"}

    with create_test_client(route_handlers=handler) as client:
        response = client.get("/")
        assert response.json() == expected


@pytest.mark.parametrize(
    "app_before_request_handler, router_before_request_handler, controller_before_request_handler, method_before_request_handler, expected",
    [
        [None, None, None, None, {"hello": "world"}],
        [sync_before_request_handler_with_return_value, None, None, None, {"hello": "moon"}],
        [None, sync_before_request_handler_with_return_value, None, None, {"hello": "moon"}],
        [None, None, sync_before_request_handler_with_return_value, None, {"hello": "moon"}],
        [None, None, None, sync_before_request_handler_with_return_value, {"hello": "moon"}],
        [
            sync_before_request_handler_with_return_value,
            async_before_request_handler_without_return_value,
            None,
            None,
            {"hello": "world"},
        ],
        [
            None,
            sync_before_request_handler_with_return_value,
            async_before_request_handler_without_return_value,
            None,
            {"hello": "world"},
        ],
        [
            None,
            None,
            sync_before_request_handler_with_return_value,
            async_before_request_handler_without_return_value,
            {"hello": "world"},
        ],
        [None, None, None, async_before_request_handler_without_return_value, {"hello": "world"}],
    ],
)
def test_before_request_handler_resolution(
    app_before_request_handler: Optional[BeforeRequestHookHandler],
    router_before_request_handler: Optional[BeforeRequestHookHandler],
    controller_before_request_handler: Optional[BeforeRequestHookHandler],
    method_before_request_handler: Optional[BeforeRequestHookHandler],
    expected: dict,
) -> None:
    class MyController(Controller):
        path = "/hello"

        before_request = controller_before_request_handler

        @get(before_request=method_before_request_handler)
        def hello(self) -> dict:
            return {"hello": "world"}

    router = Router(path="/greetings", route_handlers=[MyController], before_request=router_before_request_handler)

    with create_test_client(route_handlers=router, before_request=app_before_request_handler) as client:
        response = client.get("/greetings/hello")
        assert response.json() == expected



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from litestar.serialization import get_serializer
from litestar.utils.deprecation import warn_deprecation
from litestar.utils.scope.state import delete_litestar_scope_state as _delete_litestar_scope_state
from litestar.utils.scope.state import get_litestar_scope_state as _get_litestar_scope_state
from litestar.utils.scope.state import set_litestar_scope_state as _set_litestar_scope_state

if TYPE_CHECKING:
    from litestar.types import Scope, Serializer

__all__ = ("get_serializer_from_scope",)


def get_serializer_from_scope(scope: Scope) -> Serializer:
    """Return a serializer given a scope object.

    Args:
        scope: The ASGI connection scope.

    Returns:
        A serializer function
    """
    route_handler = scope["route_handler"]
    app = scope["app"]

    if hasattr(route_handler, "resolve_type_encoders"):
        type_encoders = route_handler.resolve_type_encoders()
    else:
        type_encoders = app.type_encoders or {}

    if response_class := (
        route_handler.resolve_response_class()  # pyright: ignore
        if hasattr(route_handler, "resolve_response_class")
        else app.response_class
    ):
        type_encoders = {**type_encoders, **(response_class.type_encoders or {})}

    return get_serializer(type_encoders)


_deprecated_names = {
    "get_litestar_scope_state": _get_litestar_scope_state,
    "set_litestar_scope_state": _set_litestar_scope_state,
    "delete_litestar_scope_state": _delete_litestar_scope_state,
}


def __getattr__(name: str) -> Any:
    if name in _deprecated_names:
        warn_deprecation(
            deprecated_name=f"litestar.utils.scope.{name}",
            version="2.4",
            kind="import",
            removal_in="3.0",
            info=f"'litestar.utils.scope.{name}' is deprecated. The Litestar scope state is private and should not be "
            f"used. Plugin authors should maintain their own scope state namespace.",
        )
        return globals()["_deprecated_names"][name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")  # pragma: no cover



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from litestar.openapi.spec.base import BaseSchemaObject

if TYPE_CHECKING:
    from litestar.openapi.spec.oauth_flow import OAuthFlow

__all__ = ("OAuthFlows",)


@dataclass
class OAuthFlows(BaseSchemaObject):
    """Allows configuration of the supported OAuth Flows."""

    implicit: OAuthFlow | None = None
    """Configuration for the OAuth Implicit flow."""

    password: OAuthFlow | None = None
    """Configuration for the OAuth Resource Owner Password flow."""

    client_credentials: OAuthFlow | None = None
    """Configuration for the OAuth Client Credentials flow. Previously called ``application`` in OpenAPI 2.0."""

    authorization_code: OAuthFlow | None = None
    """Configuration for the OAuth Authorization Code flow. Previously called ``accessCode`` in OpenAPI 2.0."""
